"""
Benchmark functions for :func:`locan.data.LocData`
"""

from copy import deepcopy

import numpy as np

import locan as lc
from locan.data.locdata import LocData

rng = np.random.default_rng(seed=1)


class BenchmarkLocDataFromSelection:
    """
    Benchmarks for selecting LocData objects
    """

    def setup(self):
        path = lc.ROOT_DIR / "tests/test_data/five_blobs.txt"
        self.locdata = lc.load_txt_file(path)
        self.df = deepcopy(self.locdata.data)
        # print(self.locdata.data.info())

        n_selected_points = 40
        self.selection_indices = rng.integers(
            low=0, high=50, size=(10, n_selected_points)
        )

        self.selection = LocData.from_selection(
            locdata=self.locdata, indices=self.selection_indices[0].tolist()
        )

        self.locdatas = [
            LocData.from_selection(locdata=self.locdata, indices=idxs.tolist())
            for idxs in self.selection_indices
        ]
        self.collection = lc.LocData.from_collection(locdatas=self.locdatas)

    def time_initialize_locdata_from_dataframe(self):
        LocData.from_dataframe(dataframe=self.df)

    def time_locdata_selection(self):
        LocData.from_selection(
            locdata=self.locdata, indices=self.selection_indices[0].tolist()
        )

    def time_locdata_collection(self):
        lc.LocData.from_collection(locdatas=self.locdatas)

    def mem_empty_locdata(self):
        return lc.LocData()

    def mem_locdata_from_dataframe(self):
        return self.locdata

    def mem_locdata_selection(self):
        return self.selection

    def mem_locdata_selection_reduce(self):
        return self.selection.reduce()

    def mem_locdata_collection(self):
        return self.collection

    def mem_locdata_collection_reduce(self):
        return self.collection.reduce()


def main():
    bm = BenchmarkLocDataFromSelection()
    bm.setup()
    bm.time_initialize_locdata_from_dataframe()
    bm.time_locdata_selection()
    bm.time_locdata_collection()


def main_profile():
    """
    Run benchmarks through profiler.
    """
    import cProfile
    import pstats

    bm = BenchmarkLocDataFromSelection()
    bm.setup()

    profiler = cProfile.Profile()
    profiler.enable()

    for _ in range(100):
        bm.time_initialize_locdata_from_dataframe()
        bm.time_locdata_selection()
        bm.time_locdata_collection()

    profiler.disable()
    stats = pstats.Stats(profiler).sort_stats("cumtime")
    stats.print_stats("time_")
    stats.print_stats()


if __name__ == "__main__":
    # main()
    main_profile()



from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class Analysis_routine(_message.Message):
    __slots__ = ("name", "parameter")
    NAME_FIELD_NUMBER: _ClassVar[int]
    PARAMETER_FIELD_NUMBER: _ClassVar[int]
    name: str
    parameter: str
    def __init__(
        self, name: _Optional[str] = ..., parameter: _Optional[str] = ...
    ) -> None: ...

class AMetadata(_message.Message):
    __slots__ = (
        "identifier",
        "comment",
        "method",
        "map",
        "creation_time",
        "modification_time",
    )

    class MapEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[str] = ...
        ) -> None: ...

    IDENTIFIER_FIELD_NUMBER: _ClassVar[int]
    COMMENT_FIELD_NUMBER: _ClassVar[int]
    METHOD_FIELD_NUMBER: _ClassVar[int]
    MAP_FIELD_NUMBER: _ClassVar[int]
    CREATION_TIME_FIELD_NUMBER: _ClassVar[int]
    MODIFICATION_TIME_FIELD_NUMBER: _ClassVar[int]
    identifier: str
    comment: str
    method: Analysis_routine
    map: _containers.ScalarMap[str, str]
    creation_time: _timestamp_pb2.Timestamp
    modification_time: _timestamp_pb2.Timestamp
    def __init__(
        self,
        identifier: _Optional[str] = ...,
        comment: _Optional[str] = ...,
        method: _Optional[_Union[Analysis_routine, _Mapping]] = ...,
        map: _Optional[_Mapping[str, str]] = ...,
        creation_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        modification_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
    ) -> None: ...



import numpy as np
import pytest

from locan.utils.statistics import (
    WeightedMeanVariance,
    biased_variance,
    ratio_fwhm_to_sigma,
    weighted_mean_variance,
)


def test_weighted_mean_variance():
    wmv = weighted_mean_variance(values=[1, 2], weights=[1, 1])
    assert wmv == WeightedMeanVariance(weighted_mean=1.5, weighted_mean_variance=0.25)

    wmv = weighted_mean_variance(values=[1, 2], weights=None)
    assert wmv == WeightedMeanVariance(weighted_mean=1.5, weighted_mean_variance=0.25)

    wmv = weighted_mean_variance(values=[1, 2], weights=[1, 0])
    assert wmv == WeightedMeanVariance(weighted_mean=1, weighted_mean_variance=0.0)

    wmv = weighted_mean_variance(values=[1, 2], weights=[1, -2])
    assert wmv == WeightedMeanVariance(weighted_mean=3, weighted_mean_variance=16.0)

    wmv = weighted_mean_variance(values=[1, 2], weights=[1, 1e100])
    assert wmv == pytest.approx((2, 0), rel=1e-6)

    wmv = weighted_mean_variance(values=[1, np.nan], weights=None)
    assert np.isnan(wmv.weighted_mean)
    assert np.isnan(wmv.weighted_mean_variance)

    wmv = weighted_mean_variance(values=[1, 2], weights=[1, np.nan])
    assert np.isnan(wmv.weighted_mean)
    assert np.isnan(wmv.weighted_mean_variance)

    with pytest.raises(TypeError):
        weighted_mean_variance(values=[1, 2], weights=[1])

    wmv = weighted_mean_variance(values=[1, 2], weights=[1, 2])
    assert wmv == pytest.approx((1.667, 0.198), rel=0.01)

    wmv = weighted_mean_variance(values=[1], weights=[2])
    assert wmv == pytest.approx((1, 0), rel=0.01)

    data = np.arange(6).reshape((3, 2))
    with pytest.raises(ValueError):
        weighted_mean_variance(values=data, weights=None)


def test_ratio_fwhm_to_sigma():
    assert ratio_fwhm_to_sigma() == pytest.approx(2.355, rel=0.001)


def test_biased_variance():
    bv = biased_variance(variance=1, n_samples=3)
    assert isinstance(bv, float)
    assert bv == pytest.approx(0.6666666666666667)

    bv = biased_variance(variance=1, n_samples=[2, 3])
    assert isinstance(bv, np.ndarray)
    assert bv == pytest.approx([0.5, 0.66666667])

    bv = biased_variance(variance=np.array([1, 2]), n_samples=3)
    assert bv == pytest.approx([0.66666667, 1.33333333])

    bv = biased_variance(variance=np.array([1, 2]), n_samples=[2, 3])
    assert bv == pytest.approx([0.5, 1.33333333])



"""
Miscellaneous functions without a home.
"""

from __future__ import annotations

from collections.abc import Iterator
from typing import Any

__all__: list[str] = ["iterate_2d_array"]


def iterate_2d_array(n_elements: int = 5, n_cols: int = 2) -> Iterator[tuple[int, int]]:
    """
    Iterator for 2-dimensional array iterating first over columns then over rows.

    Parameters
    ----------
    n_elements
        Number of elements
    n_cols
        Number of columns

    Returns
    -------
    Iterator[tuple[int, int]]
        Indices for (row, column) in each iteration.
    """
    n_rows = -(-n_elements // n_cols)
    iterator = ((i, j) for i in range(n_rows) for j in range(n_cols))
    return (item for item, _ in zip(iterator, range(n_elements)))


def _get_subclasses(cls: Any) -> set[Any]:
    """
    Recursively identify all classes that inherit from `cls`.

    Parameters
    ----------
    cls
        class object

    Returns
    -------
    set[Any]
    """
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in _get_subclasses(c)]
    )



"""

This module provides functions for rendering locdata objects in 2D.

"""

from __future__ import annotations

import logging
import warnings
from collections.abc import Callable, Sequence
from typing import TYPE_CHECKING, Any, Literal

import matplotlib.colors as mcolors
import numpy as np
import numpy.typing as npt
import scipy.signal.windows
from matplotlib import pyplot as plt

from locan.configuration import COLORMAP_DEFAULTS
from locan.data import LocData
from locan.data.locdata_utils import _check_loc_properties
from locan.dependencies import HAS_DEPENDENCY
from locan.process.aggregate import Bins, histogram
from locan.process.properties import ranges
from locan.rois.roi import Roi, _MplSelector
from locan.visualize.colormap import ColormapType, get_colormap
from locan.visualize.transform import adjust_contrast

if HAS_DEPENDENCY["mpl_scatter_density"]:
    import mpl_scatter_density

if TYPE_CHECKING:
    import boost_histogram as bh
    import matplotlib as mpl

    from locan.visualize.transform import Trafo


__all__: list[str] = [
    "render_2d_mpl",
    "render_2d_scatter_density",
    "scatter_2d_mpl",
    "apply_window",
    "render_2d_rgb_mpl",
]

logger = logging.getLogger(__name__)


def render_2d_mpl(
    locdata: LocData,
    loc_properties: list[str] | None = None,
    other_property: str | None = None,
    bins: Bins | bh.axis.Axis | bh.axis.AxesTuple | None = None,
    n_bins: int | Sequence[int] | None = None,
    bin_size: float | Sequence[float] | Sequence[Sequence[float]] | None = 10,
    bin_edges: Sequence[float] | Sequence[Sequence[float]] | None = None,
    bin_range: (
        tuple[float, float]
        | Sequence[float]
        | Sequence[Sequence[float]]
        | Literal["zero", "link"]
        | None
    ) = None,
    rescale: int | str | Trafo | Callable[..., Any] | bool | None = None,
    ax: mpl.axes.Axes | None = None,
    cmap: ColormapType = COLORMAP_DEFAULTS["CONTINUOUS"],
    cbar: bool = True,
    colorbar_kws: dict[str, Any] | None = None,
    interpolation: str = "nearest",
    **kwargs: Any,
) -> mpl.axes.Axes:
    """
    Render localization data into a 2D image by binning x,y-coordinates into
    regular bins.

    Parameters
    ----------
    locdata
        Localization data.
    loc_properties
        Localization properties to be grouped into bins.
        If None The coordinate_values of `locdata` are used.
    other_property
        Localization property (columns in `locdata.data`) that is averaged
        in each pixel.
        If None, localization counts are shown.
    bins
        The bin specification as defined in :class:`Bins`
    bin_edges
        Bin edges for all or each dimension
        with shape (dimension, n_bin_edges).
    bin_range
        Minimum and maximum edge for all or each dimensions
        with shape (2,) or (dimension, 2).
        If None (min, max) ranges are determined from data and returned;
        if 'zero' (0, max) ranges with max determined from data are returned.
        if 'link' (min_all, max_all) ranges with min and max determined from
        all combined data are returned.
    n_bins
        The number of bins for all or each dimension.
        5 yields 5 bins in all dimensions.
        (2, 5) yields 2 bins for one dimension and 5 for the other dimension.
    bin_size
        The size of bins for all or each bin and for all or each dimension
        with shape (dimension,) or (dimension, n_bins).
        5 would describe bin_size of 5 for all bins in all dimensions.
        ((2, 5),) yield bins of size (2, 5) for one dimension.
        (2, 5) yields bins of size 2 for one dimension and 5 for the other
        dimension.
        ((2, 5), (1, 3)) yields bins of size (2, 5) for one dimension and
        (1, 3) for the other dimension.
        To specify arbitrary sequence of `bin_size` use `bin_edges` instead.
    rescale
        Transformation as defined in :class:`locan.Trafo` or by
        transformation function.
        For None or False no rescaling occurs.
        Legacy behavior:
        For tuple with upper and lower bounds provided in percent,
        rescale intensity values to be within percentile of max and min
        intensities.
        For 'equal' intensity values are rescaled by histogram equalization.
    ax
        The axes on which to show the image
    cmap
        The Colormap object used to map normalized data values to RGBA colors.
    cbar
        If true draw a colorbar.
        The colobar axes is accessible using the cax property.
    colorbar_kws
        Keyword arguments for :func:`matplotlib.pyplot.colorbar`.
    interpolation
        Keyword argument for :func:`matplotlib.axes.Axes.imshow`.
    kwargs
        Other parameters passed to :func:`matplotlib.axes.Axes.imshow`.

    Returns
    -------
    matplotlib.axes.Axes
        Axes object with the image.
    """
    # Provide matplotlib.axes.Axes if not provided
    if ax is None:
        ax = plt.gca()

    # return ax if no or single point in locdata
    if len(locdata) < 2:
        if len(locdata) == 1:
            logger.warning("Locdata carries a single localization.")
        return ax

    data, bins, labels = histogram(
        locdata=locdata,
        loc_properties=loc_properties,
        other_property=other_property,
        bins=bins,
        n_bins=n_bins,
        bin_size=bin_size,
        bin_edges=bin_edges,
        bin_range=bin_range,
    )
    data = adjust_contrast(data, rescale)

    mappable = ax.imshow(
        data.T,
        **dict(
            {
                "origin": "lower",
                "extent": [*bins.bin_range[0], *bins.bin_range[1]],
                "cmap": get_colormap(colormap=cmap).matplotlib,
                "interpolation": interpolation,
            },
            **kwargs,
        ),
    )

    ax.set(title=labels[-1], xlabel=labels[0], ylabel=labels[1])

    if cbar:
        if colorbar_kws is None:
            plt.colorbar(mappable, ax=ax)
        else:
            plt.colorbar(mappable, **colorbar_kws)

    return ax


def render_2d_scatter_density(
    locdata: LocData,
    loc_properties: list[str] | None = None,
    other_property: str | None = None,
    bin_range: (
        tuple[float, float]
        | Sequence[float]
        | Sequence[Sequence[float]]
        | Literal["zero", "link"]
        | None
    ) = None,
    ax: mpl.axes.Axes | None = None,
    cmap: ColormapType = COLORMAP_DEFAULTS["CONTINUOUS"],
    cbar: bool = True,
    colorbar_kws: dict[str, Any] | None = None,
    **kwargs: Any,
) -> mpl.axes.Axes:
    """
    Render localization data into a 2D image by binning x,y-coordinates into regular bins.

    Prepare :class:`matplotlib.axes.Axes` with image.

    Note
    ----
    To rescale intensity values use norm keyword.

    Parameters
    ----------
    locdata
        Localization data.
    loc_properties
        Localization properties to be grouped into bins.
        If None The coordinate_values of `locdata` are used.
    other_property
        Localization property (columns in `locdata.data`) that is averaged
        in each pixel.
        If None, localization counts are shown.
    bin_range
        Minimum and maximum edge for all or each dimensions
        with shape (2,) or (dimension, 2).
        If None (min, max) ranges are determined from data and returned;
        if 'zero' (0, max) ranges with max determined from data are returned.
        if 'link' (min_all, max_all) ranges with min and max determined from
        all combined data are returned.
    ax
        The axes on which to show the image
    cmap
        The Colormap object used to map normalized data values to RGBA colors.
    cbar
        If true draw a colorbar.
        The colobar axes is accessible using the cax property.
    colorbar_kws
        Keyword arguments for :func:`matplotlib.pyplot.colorbar`.
    kwargs
        Other parameters passed to :class:`mpl_scatter_density.ScatterDensityArtist`.

    Returns
    -------
    matplotlib.axes.Axes
        Axes object with the image.

    Warnings
    ________
    This function is deprecated.
    Use third-party methods like datashader instead.
    """
    if not HAS_DEPENDENCY["mpl_scatter_density"]:
        raise ImportError("mpl-scatter-density is required.")

    warnings.warn(
        "This function is deprecated. "
        "Use third-party methods like datashader instead.",
        DeprecationWarning,
        stacklevel=2,
    )

    # Provide matplotlib.axes.Axes if not provided
    if ax is None:
        ax = plt.gca()

    # return ax if no or single point in locdata
    if len(locdata) < 2:
        if len(locdata) == 1:
            logger.warning("Locdata carries a single localization.")
        return ax
    else:
        fig = ax.get_figure()
        ax = fig.add_subplot(  # type: ignore[union-attr]
            1, 1, 1, projection="scatter_density", label="scatter_density"
        )

    if loc_properties is None:
        data = locdata.coordinates.T
        labels = list(locdata.coordinate_keys)
    elif isinstance(loc_properties, str) and loc_properties in locdata.coordinate_keys:
        data = locdata.data[loc_properties].values.T
        labels = list(loc_properties)
    elif isinstance(loc_properties, (list, tuple)):
        for prop in loc_properties:
            if prop not in locdata.coordinate_keys:
                raise ValueError(f"{prop} is not a valid property in locdata.")
        data = locdata.data[list(loc_properties)].values.T
        labels = list(loc_properties)
    else:
        raise ValueError(f"{loc_properties} is not a valid property in locdata.")

    if bin_range is None or isinstance(bin_range, str):
        bin_range_: npt.NDArray[np.float64] = ranges(locdata, loc_properties=labels, special=bin_range)  # type: ignore
    else:
        bin_range_ = bin_range  # type: ignore[assignment]

    if other_property is None:
        # histogram data by counting points
        if data.shape[0] == 2:
            values = None
        else:
            raise TypeError("Only 2D data is supported.")
        labels.append("counts")
    elif other_property in locdata.data.columns:
        # histogram data by averaging values
        if data.shape[0] == 2:
            # here color serves as weight since it is averaged over all points before binning.
            values = locdata.data[other_property].values.T  # type: ignore
        else:
            raise TypeError("Only 2D data is supported.")
        labels.append(other_property)
    else:
        raise TypeError(f"No valid property name {other_property}.")

    a = mpl_scatter_density.ScatterDensityArtist(
        ax,
        *data,
        c=values,
        origin="lower",
        extent=[*bin_range_[0], *bin_range_[1]],
        cmap=get_colormap(colormap=cmap).matplotlib,
        **kwargs,
    )
    mappable = ax.add_artist(a)
    ax.set_xlim(*bin_range_[0])
    ax.set_ylim(*bin_range_[1])

    ax.set(title=labels[-1], xlabel=labels[0], ylabel=labels[1])

    if cbar:
        if colorbar_kws is None:
            plt.colorbar(mappable, ax=ax, label=labels[-1])  # type:ignore[arg-type]
        else:
            plt.colorbar(mappable, **colorbar_kws)  # type:ignore[arg-type]

    return ax


def scatter_2d_mpl(
    locdata: LocData,
    ax: mpl.axes.Axes | None = None,
    index: bool = True,
    text_kwargs: dict[str, Any] | None = None,
    **kwargs: Any,
) -> mpl.axes.Axes:
    """
    Scatter plot of locdata elements with text marker for each element.

    Parameters
    ----------
    locdata
       Localization data.
    ax
       The axes on which to show the plot
    index
       Flag indicating if element indices are shown.
    text_kwargs
       Keyword arguments for :func:`matplotlib.axes.Axes.text`.
    kwargs
       Other parameters passed to :func:`matplotlib.axes.Axes.scatter`.

    Returns
    -------
    matplotlib.axes.Axes
       Axes object with the image.
    """
    if text_kwargs is None:
        text_kwargs = {}

    # Provide matplotlib.axes.Axes if not provided
    if ax is None:
        ax = plt.gca()

    # return ax if no or single point in locdata
    if len(locdata) < 2:
        if len(locdata) == 1:
            logger.warning("Locdata carries a single localization.")
        return ax

    coordinates = locdata.coordinates
    ax.scatter(*coordinates.T, **dict({"marker": "+", "color": "grey"}, **kwargs))

    # plot element number
    if index:
        for centroid, marker in zip(coordinates, locdata.data.index.values):
            ax.text(  # type: ignore
                *centroid, marker, **dict({"color": "grey", "size": 20}, **text_kwargs)
            )

    ax.set(xlabel="position_x", ylabel="position_y")

    return ax


def apply_window(
    image: npt.ArrayLike, window_function: str = "tukey", **kwargs: Any
) -> npt.NDArray[np.float64]:
    """
    Apply window function to image.

    Parameters
    ----------
    image
        Image
    window_function
        Window function to apply. One of 'tukey', 'hann' or any other in `scipy.signal.windows`.
    kwargs
        Other parameters passed to the `scipy.signal.windows` window function.
    """
    image = np.asarray(image)
    window_func = getattr(scipy.signal.windows, window_function)
    windows = [window_func(M, **kwargs) for M in image.shape]

    result = image.astype("float64")
    result *= windows[0]
    result *= windows[1][:, None]

    return result


def select_by_drawing_mpl(
    locdata: LocData, region_type: str = "rectangle", **kwargs: Any
) -> list[Roi]:
    """
    Select region of interest from rendered image by drawing rois.

    Parameters
    ----------
    locdata
        The localization data from which to select localization data.
    region_type
        rectangle, or ellipse specifying the selection widget to use.
    kwargs
        Other parameters as specified for :func:`render_2d`.

    Returns
    -------
    list[Roi]

    See Also
    --------
    :func:`locan.scripts.sc_draw_roi_mpl` : script for drawing rois
    matplotlib.widgets : selector functions
    """

    _fig, ax = plt.subplots(nrows=1, ncols=1)  # type: ignore
    render_2d_mpl(locdata, ax=ax, **kwargs)
    selector = _MplSelector(ax, type=region_type)
    plt.show()  # type: ignore
    roi_list = [Roi(reference=locdata, region=roi["region"]) for roi in selector.rois]
    return roi_list


def render_2d_rgb_mpl(
    locdatas: list[LocData],
    loc_properties: list[str] | None = None,
    other_property: str | None = None,
    bins: Bins | bh.axis.Axis | bh.axis.AxesTuple | None = None,
    n_bins: int | Sequence[int] | None = None,
    bin_size: float | Sequence[float] | Sequence[Sequence[float]] | None = 10,
    bin_edges: Sequence[float] | Sequence[Sequence[float]] | None = None,
    bin_range: (
        tuple[float, float]
        | Sequence[float]
        | Sequence[Sequence[float]]
        | Literal["zero", "link"]
        | None
    ) = None,
    rescale: int | str | Trafo | Callable[..., Any] | bool | None = None,
    ax: mpl.axes.Axes | None = None,
    interpolation: str = "nearest",
    **kwargs: Any,
) -> mpl.axes.Axes:
    """
    Render localization data into a 2D RGB image by binning x,y-coordinates into regular bins.

    Note
    ----
    For rescale=False no normalization is carried out image intensities are clipped to (0, 1) for float value
    or (0, 255) for integer values according to the matplotlib.imshow behavior.
    For rescale=None we apply a normalization to (min, max) of all intensity values.
    For all other rescale options the normalization is applied to each individual image.

    Parameters
    ----------
    locdatas
        Localization data.
    loc_properties
        Localization properties to be grouped into bins. If None The coordinate_values of `locdata` are used.
    other_property
        Localization property (columns in `locdata.data`) that is averaged in each pixel.
        If None, localization counts are shown.
    bins
        The bin specification as defined in :class:`Bins`
    bin_edges
        Bin edges for all or each dimension
        with shape (dimension, n_bin_edges).
    bin_range
        Minimum and maximum edge for all or each dimensions
        with shape (2,) or (dimension, 2).
        If None (min, max) ranges are determined from data and returned;
        if 'zero' (0, max) ranges with max determined from data are returned.
        if 'link' (min_all, max_all) ranges with min and max determined from
        all combined data are returned.
    n_bins
        The number of bins for all or each dimension.
        5 yields 5 bins in all dimensions.
        (2, 5) yields 2 bins for one dimension and 5 for the other dimension.
    bin_size
        The size of bins for all or each bin and for all or each dimension
        with shape (dimension,) or (dimension, n_bins).
        5 would describe bin_size of 5 for all bins in all dimensions.
        ((2, 5),) yield bins of size (2, 5) for one dimension.
        (2, 5) yields bins of size 2 for one dimension and 5 for the other
        dimension.
        ((2, 5), (1, 3)) yields bins of size (2, 5) for one dimension and
        (1, 3) for the other dimension.
        To specify arbitrary sequence of `bin_size` use `bin_edges` instead.
    rescale
        Transformation as defined in :class:`locan.Trafo` or by
        transformation function.
        For None or False no rescaling occurs.
        Legacy behavior:
        For tuple with upper and lower bounds provided in percent,
        rescale intensity values to be within percentile of max and min
        intensities.
        For 'equal' intensity values are rescaled by histogram equalization.
    ax
        The axes on which to show the image
    interpolation
        Keyword argument for :func:`matplotlib.axes.Axes.imshow`.
    kwargs
        Other parameters passed to :func:`matplotlib.axes.Axes.imshow`.

    Returns
    -------
    matplotlib.axes.Axes
        Axes object with the image.
    """
    # Provide matplotlib.axes.Axes if not provided
    if ax is None:
        ax = plt.gca()

    locdata_temp = LocData.concat(locdatas)

    # return ax if no or single point in locdata
    if len(locdata_temp) < 2:
        if len(locdata_temp) == 1:
            logger.warning("Locdata carries a single localization.")
        return ax

    if bin_edges is None:
        _, bins, labels = histogram(
            locdata=locdata_temp,
            loc_properties=loc_properties,
            other_property=other_property,
            bins=bins,
            n_bins=n_bins,
            bin_size=bin_size,
            bin_edges=bin_edges,
            bin_range=bin_range,
        )
    else:
        labels = _check_loc_properties(locdata_temp, loc_properties)
        bins = Bins(bin_edges=bin_edges, labels=labels)

    imgs = [
        histogram(
            locdata=locdata, loc_properties=loc_properties, other_property=other_property, bin_edges=bins.bin_edges  # type: ignore
        ).data
        for locdata in locdatas
    ]

    if rescale is None:
        norm: int | str | Trafo | Callable[..., Any] = mcolors.Normalize(
            vmin=np.min(imgs), vmax=np.max(imgs)
        )
    else:
        norm = rescale
    imgs = [adjust_contrast(img, rescale=norm) for img in imgs]

    new = np.zeros_like(imgs[0])
    rgb_stack = np.stack([new] * 3, axis=2)

    for i, img in enumerate(imgs):
        rgb_stack[:, :, i] = img

    rgb_stack = np.transpose(rgb_stack, axes=(1, 0, 2))
    ax.imshow(
        rgb_stack,
        **dict(
            {
                "origin": "lower",
                "extent": [*bins.bin_range[0], *bins.bin_range[1]],
                "interpolation": interpolation,
            },
            **kwargs,
        ),
    )

    ax.set(title=labels[-1], xlabel=labels[0], ylabel=labels[1])

    return ax



"""

Compute statistics for localization data.

These values can represent new properties of locdata.

"""

from __future__ import annotations

from collections import namedtuple
from collections.abc import Iterable
from typing import Any, Literal

import numpy as np
import numpy.typing as npt
import pandas as pd

from locan.data.locdata import LocData

__all__: list[str] = ["statistics", "ranges", "range_from_collection"]


def statistics(
    locdata: LocData | pd.DataFrame | pd.Series[Any],
    statistic_keys: str | Iterable[str] = (
        "count",
        "min",
        "max",
        "mean",
        "median",
        "std",
        "sem",
    ),
) -> dict[str, Any]:
    """
    Compute selected statistical parameter for localization data.

    Parameters
    ----------
    locdata
        Localization data

    statistic_keys
        Pandas statistic functions.
        Default: ('count', 'min', 'max', 'mean', 'median', 'std', 'sem')

    Returns
    -------
    dict[str, Any]
        A dict with descriptive statistics.
    """
    data: pd.DataFrame | pd.Series[Any]
    if isinstance(locdata, LocData):
        data = locdata.data
    elif isinstance(locdata, (pd.DataFrame, pd.Series)):
        data = locdata
    else:
        raise TypeError("locdata should be of type locan.LocData or pandas.DataFrame.")

    statistics_ = data.agg(statistic_keys)  # type: ignore

    if isinstance(locdata, pd.Series):
        p = str(data.name)
        if isinstance(statistic_keys, str):
            dict_ = {p + "_" + statistic_keys: statistics_}
        else:
            dict_ = {p + "_" + s: statistics_[s] for s in statistic_keys}  # type: ignore
    else:
        if isinstance(statistic_keys, str):
            generator = (p for p in list(data))
            dict_ = {p + "_" + statistic_keys: statistics_[p] for p in generator}  # type: ignore
        else:
            generator = ((p, s) for p in list(data) for s in statistic_keys)  # type: ignore
            dict_ = {p + "_" + s: statistics_[p][s] for p, s in generator}  # type: ignore

    return dict_


def ranges(
    locdata: LocData,
    loc_properties: str | Iterable[str] | Literal[True] | None = None,
    special: Literal["zero", "link"] | None = None,
    epsilon: float = 1,
) -> npt.NDArray[np.float64] | None:
    """
    Provide data ranges for localization properties.
    If LocData is empty, None is returned.
    If LocData carries a single localization, the range will be
    (value, value + `epsilon`).

    Parameters
    ----------
    locdata
        Localization data.
    loc_properties
        Localization properties for which the range is determined.
        If None the ranges for all spatial coordinates are returned.
        If True the ranges for all localization properties are returned.
    special
        If None (min, max) ranges are determined from data and returned;
        if 'zero' (0, max) ranges with max determined from data are returned.
        if 'link' (min_all, max_all) ranges with min and max determined from
        all combined data are returned.
    epsilon : float
        number to specify the range for single values in locdata.

    Returns
    -------
    npt.NDArray[np.float64] | None
        The data range (min, max) for each localization property.
        Array of shape (dimension, 2).
    """
    if locdata.data.empty:
        return None
    elif len(locdata) == 1:
        pass

    if loc_properties is None:
        ranges_ = locdata.bounding_box.hull.T.copy()  # type: ignore
    elif loc_properties is True:
        ranges_ = np.array([locdata.data.min(), locdata.data.max()]).T
    elif isinstance(loc_properties, str):
        ranges_ = np.array(
            [[locdata.data[loc_properties].min(), locdata.data[loc_properties].max()]]
        )
    else:
        loc_properties = list(loc_properties)
        ranges_ = np.array(
            [locdata.data[loc_properties].min(), locdata.data[loc_properties].max()]
        ).T

    if len(locdata) == 1:
        if ranges_.size == 0:
            ranges_ = np.concatenate(
                [locdata.coordinates, locdata.coordinates + epsilon], axis=0
            ).T
        else:
            ranges_ = ranges_ + [0, epsilon]

    if special is None:
        pass
    elif special == "zero":
        ranges_[:, 0] = 0
    elif special == "link":
        minmax = np.array([ranges_[:, 0].min(axis=0), ranges_[:, 1].max(axis=0)])
        ranges_ = np.repeat(minmax[None, :], len(ranges_), axis=0)
    else:
        raise ValueError(f"The parameter special={special} is not defined.")

    return ranges_


def range_from_collection(
    locdatas: list[LocData],
    loc_properties: str | Iterable[str] | Literal[True] | None = None,
    special: Literal["zero", "link"] | None = None,
    epsilon: float = 1,
) -> tuple[tuple[float, float], ...]:
    """
    Compute the maximum range from all combined localizations for each
    dimension.

    Parameters
    ----------
    locdatas
        Collection of localization datasets.
    loc_properties
        Localization properties for which the range is determined.
        If None the ranges for all spatial coordinates are returned.
        If True the ranges for all locdata.data properties are returned.
    special
        If None (min, max) ranges are determined from data and returned;
        if 'zero' (0, max) ranges with max determined from data are returned.
        if 'link' (min_all, max_all) ranges with min and max determined from
        all combined data are returned.
    epsilon
        number to specify the range for single values in locdata.

    Returns
    -------
    namedtuple
        A namedtuple('Ranges', locdata.coordinate_keys)
        of namedtuple('Range', 'min max').
    """
    ranges_ = [
        ranges(
            locdata=locdata,
            loc_properties=loc_properties,
            special=special,
            epsilon=epsilon,
        )
        for locdata in locdatas
    ]

    mins = np.min([rand[:, 0] for rand in ranges_ if rand is not None], axis=0)
    maxs = np.max([rand[:, 1] for rand in ranges_ if rand is not None], axis=0)

    if loc_properties is None:
        labels = locdatas[0].coordinate_keys
    elif loc_properties is True:
        labels = locdatas[0].data.columns  # type: ignore
    elif isinstance(loc_properties, str):
        labels = [loc_properties]
    else:
        labels = loc_properties  # type: ignore

    Ranges = namedtuple("Ranges", labels)  # type: ignore[misc]
    Range = namedtuple("Range", "min max")
    result = Ranges(
        *(Range(min_value, max_value) for min_value, max_value in zip(mins, maxs))
    )
    return result



import numpy as np
import pandas as pd
import pytest

from locan import LocData
from locan.dependencies import HAS_DEPENDENCY

if HAS_DEPENDENCY["trackpy"]:
    from trackpy import quiet as tp_quiet

    from locan.process.tracking import link_locdata, track


pytestmark = pytest.mark.skipif(
    not HAS_DEPENDENCY["trackpy"], reason="requires trackpy"
)

if HAS_DEPENDENCY["trackpy"]:
    tp_quiet()  # same as: trackpy.logger.setLevel(logging.WARN)


@pytest.fixture()
def locdata_simple():
    dict_ = {
        "position_x": [0, 1, 2, 10, 20, 21, 30, 4],
        "position_y": [0, 1, 2, 10, 20, 21, 30, 4],
        "position_z": [0, 1, 2, 10, 20, 21, 30, 4],
        "frame": np.arange(8),
    }
    return LocData(dataframe=pd.DataFrame.from_dict(dict_))


def test_link_locdata(locdata_simple):
    track_series = link_locdata(locdata_simple, search_range=5, memory=0)
    assert len(track_series) == 8
    assert track_series.name == "track"


def test_track(locdata_simple):
    locdata_new, _track_series = track(locdata_simple, search_range=5)
    expected = {
        "localization_count": 3,
        "position_x": 1.0,
        "uncertainty_x": 0.5773502691896257,
        "position_y": 1.0,
        "uncertainty_y": 0.5773502691896257,
        "position_z": 1.0,
        "uncertainty_z": 0.5773502691896257,
        "frame": 0,
        "region_measure_bb": 8,
        "localization_density_bb": 0.375,
        "subregion_measure_bb": 12,
    }
    for value, value_expected, values_in_data in zip(
        locdata_new.references[0].properties.values(),
        expected.values(),
        locdata_new.data.iloc[0].to_numpy(),
    ):
        assert value == value_expected == values_in_data

    expected = {
        "localization_count": 1,
        "position_x": 10,
        "uncertainty_x": 0,
        "position_y": 10,
        "uncertainty_y": 0,
        "position_z": 10,
        "uncertainty_z": 0,
        "frame": 3,
        "region_measure_bb": 0,
    }
    for value, value_expected, values_in_data in zip(
        locdata_new.references[1].properties.values(),
        expected.values(),
        locdata_new.data.iloc[1].to_numpy(),
    ):
        assert value == value_expected == values_in_data

    assert "frame" in locdata_new.data.columns
    assert len(locdata_new) == 5

    locdata_new, _track_series = track(locdata_simple, search_range=5, memory=5)
    assert len(locdata_new) == 4



"""
Compute localization uncertainty.

A theoretical estimate for localization uncertainty is given by the
Cramer-Rao-lower-bound for the localization precision.
The localization precision depends on a number of experimental factors
including camera and photophysical characteristics [1]_ [2]_.
We provide functions to compute localization uncertainty from available
localization properties as predicted by the Cramer-Rao-lower-bound for the
localization precision.

References
----------
.. [1] K.I. Mortensen, L. S. Churchman, J. A. Spudich, H. Flyvbjerg,
   Nat. Methods 7 (2010): 377–384.
.. [2] Rieger B., Stallinga S.,
   The lateral and axial localization uncertainty in super-resolution
   light microscopy.
   Chemphyschem 17;15(4), 2014:664-70. doi: 10.1002/cphc.201300711
"""

from __future__ import annotations

import logging
import sys
from collections.abc import Callable
from inspect import signature
from typing import TYPE_CHECKING, Any, cast

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

if TYPE_CHECKING:
    from locan.data.locdata import LocData

import numpy as np
import numpy.typing as npt
import pandas as pd

from locan.analysis import metadata_analysis_pb2
from locan.analysis.analysis_base import _Analysis
from locan.data.locdata_utils import _get_loc_property_key_per_dimension

__all__: list[str] = ["LocalizationUncertainty"]

logger = logging.getLogger(__name__)

LocalizationPrecisionModel = Callable[..., npt.NDArray[np.float64]]


# The algorithms


def localization_precision_model_1(intensity: npt.ArrayLike) -> npt.NDArray[np.float64]:
    """
    Localization precision as function of
    intensity (I) according to:

    .. math::

        \\sigma_{loc} = \\frac{1}{\\sqrt{I}}


    Parameters
    ----------
    intensity
        Intensity values

    Returns
    -------
    npt.NDArray[np.float64]
    """
    intensity = np.asarray(intensity)
    sigma: npt.NDArray[np.float64] = 1 / np.sqrt(intensity)
    return sigma


def localization_precision_model_2(
    intensity: npt.ArrayLike, psf_sigma: npt.ArrayLike
) -> npt.NDArray[np.float64]:
    """
    Localization precision as function of
    intensity (I), PSF width (sigma_PSF) according to:

    .. math::

        \\sigma_{loc} = \\frac{\\sigma_{PSF}}{\\sqrt{I}}

    Parameters
    ----------
    intensity
        Intensity values
    psf_sigma
        The PSF size

    Returns
    -------
    npt.NDArray[np.float64]
    """
    intensity = np.asarray(intensity)
    psf_sigma = np.asarray(psf_sigma)
    sigma: npt.NDArray[np.float64] = psf_sigma / np.sqrt(intensity)
    return sigma


def localization_precision_model_3(
    intensity: npt.ArrayLike,
    psf_sigma: npt.ArrayLike,
    pixel_size: npt.ArrayLike,
    local_background: npt.ArrayLike,
) -> npt.NDArray[np.float64]:
    """
    Localization precision as function of
    intensity (I), PSF size (\\sigma_{PSF}), pixel size (a), background (b)
    according to:

    .. math::

        \\sigma_{a} ^{2} = \\sigma_{PSF} ^{2} + a^{2} / 12

        \\tau = 2 * \\pi * b * \\sigma_{a} ^{2} / (I * a^{2})

        \\sigma_{loc} ^{2} = \\sigma_{a} ^{2} / I * (1 + 4 * \\tau + \\sqrt{2 * \\tau / (1 + 4 * \\tau) } )


    Parameters
    ----------
    intensity
        Intensity values
    psf_sigma
        The PSF size
    pixel_size
        Size of camera pixel
    local_background
        The local background

    Returns
    -------
    npt.NDArray[np.float64]
    """
    intensity = np.asarray(intensity)
    psf_sigma = np.asarray(psf_sigma)
    pixel_size = np.asarray(pixel_size)
    local_background = np.asarray(local_background)

    sigma_a_squared = psf_sigma**2 + pixel_size**2 / 12
    tau = 2 * np.pi * sigma_a_squared * local_background / (intensity * pixel_size**2)

    sigma_squared = (
        sigma_a_squared / intensity * (1 + 4 * tau + np.sqrt(2 * tau / (1 + 4 * tau)))
    )
    sigma: npt.NDArray[np.float64] = np.sqrt(sigma_squared)
    return sigma


def _localization_uncertainty(
    locdata: LocData, model: int | LocalizationPrecisionModel, **kwargs: Any
) -> pd.DataFrame:
    if callable(model):
        model = cast(Callable, model)  # type: ignore[type-arg]
    elif model == 1:
        model = cast(Callable, model)  # type: ignore[type-arg]
        model = localization_precision_model_1
    elif model == 2:
        model = cast(Callable, model)  # type: ignore[type-arg]
        model = localization_precision_model_2
    elif model == 3:
        model = cast(Callable, model)  # type: ignore[type-arg]
        model = localization_precision_model_3
    else:
        raise TypeError("model must be 1, 2, 3 or callable.")

    params = signature(model).parameters.keys()

    for key_ in kwargs.keys():
        if key_ not in params:
            if key_[:-2] not in params:
                raise KeyError(f"Kwarg {key_} does not fit the models signature.")

    # todo: improve checks on localization_property units
    try:
        if (
            not [
                prop_.unit
                for prop_ in locdata.meta.localization_properties
                if prop_.name == "intensity"
            ][0]
            == "photons"
        ):
            raise UserWarning
    except (UserWarning, IndexError):
        logger.warning(
            "The localization property `intensity` does not have the unit photons."
        )
    # todo: check if unit psf_sigma and coordinates is the same.

    available_keys_list = [
        _get_loc_property_key_per_dimension(locdata=locdata.data, property_key=param_)
        for param_ in params
    ]
    available_coordinate_keys = _get_loc_property_key_per_dimension(
        locdata=locdata.data, property_key="position"
    )
    new_uncertainty_keys = ["uncertainty_x", "uncertainty_y", "uncertainty_z"]

    results_dict = dict()
    for i, (c_key, new_u_key) in enumerate(
        zip(available_coordinate_keys, new_uncertainty_keys)
    ):
        # transpose available_keys_list
        available_keys = [item[i] for item in available_keys_list]
        if c_key is not None:
            suffix = "_" + c_key.split("_")[-1]
            args = []
            # go through all args for model
            for key_, param_ in zip(available_keys, params):  # type: ignore
                if key_ in kwargs.keys():
                    args.append(kwargs[key_])
                elif key_ is None and param_ in kwargs.keys():
                    args.append(kwargs[param_])
                elif key_ is None and param_ + suffix in kwargs.keys():
                    args.append(kwargs[param_ + suffix])
                elif key_ is None:
                    args.append(None)
                    break
                else:
                    args.append(locdata.data[key_])
            if all(arg_ is not None for arg_ in args):
                results_dict[new_u_key] = model(*args)

    return pd.DataFrame(results_dict)


# The specific analysis classes


class LocalizationUncertainty(_Analysis):
    """
    Compute the Cramer Rao lower bound for localization uncertainty
    for each localization's spatial coordinate in locdata.

    Uncertainty is computed according to one of the following model functions:

    1) 1 / sqrt(intensity)
    2) psf_sigma / sqrt(intensity)
    3) f(psf_sigma, intensity, pixel_size, background)

    Localization properties have to be available for all or each spatial
    dimension (like `psf_sigma` or `psf_sigma_x`).
    The localization property `intensity` must have the unit `photons`.
    The unit of `pixel_size` must be the same as that of position coordinates.

    Parameters
    ----------
    meta : locan.analysis.metadata_analysis_pb2.AMetadata | None
        Metadata about the current analysis routine.
    model : int
        Model function for theoretical localization uncertainty.
    kwargs
        kwargs for the chosen model.
        If none are given the localization properties are taken from locdata.


    Attributes
    ----------
    count : int
        A counter for counting instantiations.
    parameter : dict
        A dictionary with all settings for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    results : pandas.DataFrame
        Uncertainty for each localization and in each dimension.
    """

    count = 0

    def __init__(
        self,
        meta: metadata_analysis_pb2.AMetadata | None = None,
        model: int = 1,
        **kwargs: Any,
    ) -> None:
        parameters = self._get_parameters(locals())
        super().__init__(**parameters)
        self.results = None

    def compute(self, locdata: LocData) -> Self:
        """
        Run the computation.

        Parameters
        ----------
        locdata
            Localization data.

        Returns
        -------
        Self
        """
        if not len(locdata):
            logger.warning("Locdata is empty.")
            return self

        self.results = _localization_uncertainty(locdata=locdata, **self.parameter)
        return self



"""

Analyze grouped property expectations.

As an example, assume a collection of locdatas representing localization
clusters.
Some property of locdata `loc_property`(i.e. cluster property; e.g. intensity)
might depend on another property `other_loc_property` (e.g. number of
localizations in cluster).
For inspection, all values of loc_property are grouped by other_loc_property
and mean with standard deviations are displayed.

See Also
--------
position_variance_expectation
convex_hull_expectation
"""

from __future__ import annotations

import logging
import sys
from collections.abc import Mapping, Sequence
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

import boost_histogram as bh
import matplotlib.pyplot as plt
import pandas as pd

from locan.analysis import metadata_analysis_pb2
from locan.analysis.analysis_base import _Analysis
from locan.process.aggregate import Bins

if TYPE_CHECKING:
    import matplotlib as mpl

    from locan.data.locdata import LocData

__all__: list[str] = ["GroupedPropertyExpectation"]

logger = logging.getLogger(__name__)


@dataclass(repr=False)
class GroupedPropertyExpectationResults:
    values: pd.DataFrame = field(default_factory=pd.DataFrame)
    # with index being reference_index
    # with columns: loc_property, other_loc_property, value_to_expectation_ratio
    grouped: pd.DataFrame = field(default_factory=pd.DataFrame)
    # with index being other_loc_property
    # with columns: loc_property_mean, loc_property_std, expectation


class GroupedPropertyExpectation(_Analysis):
    """
    Analyze variation of localization property
    in relation to other localization property that is grouped.

    Parameters
    ----------
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    loc_property: str | None
        The localization property to analyze.
    other_loc_property: str | None
        The localization property to group on.
    expectation : int | float | Mapping | pd.Series[Any] | None
        The expected value for all or each other localization property.

    Attributes
    ----------
    count : int
        A counter for counting instantiations (class attribute).
    parameter : dict[str, Any]
        A dictionary with all settings for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    results : GroupedPropertyExpectationResults
        Computed results.
    distribution_statistics : Distribution_stats, None
        Distribution parameters derived from MLE fitting of results.grouped.
    """

    def __init__(
        self,
        meta: metadata_analysis_pb2.AMetadata | None = None,
        loc_property: str | None = None,
        other_loc_property: str | None = None,
        expectation: int | float | Mapping[str, Any] | pd.Series[Any] | None = None,
    ) -> None:
        parameters = self._get_parameters(locals())
        super().__init__(**parameters)
        self.expectation = expectation
        self.results = None
        self.distribution_statistics = None

    def compute(self, locdata: LocData) -> Self:
        """
        Run the computation.

        Parameters
        ----------
        locdata
            Localization data.

        Returns
        -------
        Self
        """
        if not len(locdata):
            logger.warning("Locdata is empty.")
            return self

        loc_property = self.parameter["loc_property"]
        other_loc_property = self.parameter["other_loc_property"]

        self.results = GroupedPropertyExpectationResults()
        self.results.values = locdata.data.loc[:, [loc_property, other_loc_property]]

        grouped = self.results.values.groupby(other_loc_property)

        self.results.grouped[loc_property + "_mean"] = grouped.mean()
        self.results.grouped[loc_property + "_std"] = grouped.std()

        if self.expectation is None:
            self.results.grouped["expectation"] = pd.NA
        elif isinstance(self.expectation, (int, float)):
            self.results.grouped["expectation"] = self.expectation
        elif isinstance(self.expectation, pd.Series):
            indices = self.results.grouped.index
            self.results.grouped["expectation"] = self.expectation.loc[indices]
        elif isinstance(self.expectation, Mapping):
            self.results.grouped["expectation"] = [
                self.expectation[index_] for index_ in self.results.grouped.index
            ]

        self.results.values["expectation"] = self.results.grouped.loc[
            self.results.values[other_loc_property], "expectation"
        ].to_numpy()

        self.results.values["value_to_expectation_ratio"] = (
            self.results.values[loc_property] / self.results.values["expectation"]
        )

        return self

    def plot(self, ax: mpl.axes.Axes | None = None, **kwargs: Any) -> mpl.axes.Axes:
        """
        Provide plot as :class:`matplotlib.axes.Axes` object showing the
        variances as function of localization counts.

        Parameters
        ----------
        ax
            The axes on which to show the image
        kwargs
            Other parameters passed to :func:`matplotlib.pyplot.plot`.

        Returns
        -------
        matplotlib.axes.Axes
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if self.results is None:
            return ax

        self.results.values.plot(
            kind="scatter",
            alpha=0.2,
            x=self.parameter["other_loc_property"],
            y=self.parameter["loc_property"],
            color="gray",
            ax=ax,
            **dict(dict(legend=False), **kwargs),
        )

        self.results.grouped.plot(
            kind="line",
            marker=".",
            x=None,
            y=self.parameter["loc_property"] + "_mean",
            color="blue",
            ax=ax,
        )

        if not self.results.grouped.expectation.isna().all():
            self.results.grouped.plot(
                kind="line",
                x=None,
                y="expectation",
                color="black",
                ax=ax,
            )

        ax.set(
            title="Grouped Property Expectation",
            xlabel=self.parameter["other_loc_property"],
            ylabel=self.parameter["loc_property"],
        )

        return ax

    def hist(
        self,
        ax: mpl.axes.Axes | None = None,
        bins: Bins | bh.axis.Axis | bh.axis.AxesTuple | None = None,
        n_bins: int | Sequence[int] | None = None,
        bin_size: float | Sequence[float] | Sequence[Sequence[float]] | None = None,
        bin_edges: Sequence[float] | Sequence[Sequence[float]] | None = None,
        bin_range: (
            tuple[float, float] | Sequence[float] | Sequence[Sequence[float]] | None
        ) = None,
        log: bool = True,
        fit: bool = False,
        **kwargs: Any,
    ) -> mpl.axes.Axes:
        """
        Provide plot as :class:`matplotlib.axes.Axes` object showing the
        2-dimensional histogram.

        Parameters
        ----------
        ax
            The axes on which to show the image
        bins
            The bin specification as defined in :class:`Bins`
        bin_edges
            Bin edges for all or each dimension
            with shape (dimension, n_bin_edges).
        bin_range
            Minimum and maximum edge for all or each dimensions
            with shape (2,) or (dimension, 2).
        n_bins
            The number of bins for all or each dimension.
            5 yields 5 bins in all dimensions.
            (2, 5) yields 2 bins for one dimension and 5 for the other dimension.
        bin_size
            The size of bins for all or each bin and for all or each dimension
            with shape (dimension,) or (dimension, n_bins).
            5 would describe bin_size of 5 for all bins in all dimensions.
            ((2, 5),) yield bins of size (2, 5) for one dimension.
            (2, 5) yields bins of size 2 for one dimension and 5 for the other
            dimension.
            ((2, 5), (1, 3)) yields bins of size (2, 5) for one dimension and
            (1, 3) for the other dimension.
            To specify arbitrary sequence of `bin_size` use `bin_edges` instead.
        log
            Flag for plotting on a log scale.
        fit
            Flag indicating if distribution fit is shown.
        kwargs
            Other parameters passed to :func:`matplotlib.pyplot.pcolormesh`.

        Returns
        -------
        matplotlib.axes.Axes
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if self.results is None:
            return ax

        fig = ax.get_figure()

        other_loc_property = self.parameter["other_loc_property"]
        loc_property = self.parameter["loc_property"]

        if all(
            item_ is None for item_ in [bins, n_bins, bin_size, bin_edges, bin_range]
        ):
            max_other_loc_property_ = self.results.grouped.index.max()
            max_loc_property_ = self.results.values[loc_property].max()

            if log:
                axes = bh.axis.AxesTuple(
                    (
                        bh.axis.Regular(
                            max_other_loc_property_,
                            1,
                            max_other_loc_property_,
                            transform=bh.axis.transform.log,
                        ),
                        bh.axis.Regular(
                            50, 1, max_loc_property_, transform=bh.axis.transform.log
                        ),
                    )
                )
            else:
                axes = bh.axis.AxesTuple(
                    (
                        bh.axis.Regular(
                            max_other_loc_property_, 1, max_other_loc_property_
                        ),
                        bh.axis.Regular(50, 1, max_loc_property_),
                    )
                )
            bins = Bins(bins=axes)
        else:
            try:
                bins = Bins(
                    bins,
                    n_bins,
                    bin_size,
                    bin_edges,
                    bin_range,
                    labels=[loc_property, other_loc_property],
                )
            except ValueError as exc:
                # todo: check if message is appropriate
                # the error is raised again only to adapt the message.
                raise ValueError(
                    "Bin dimension and len of `loc_properties` is incompatible."
                ) from exc
        axes = bins.boost_histogram_axes
        histogram = bh.Histogram(*axes)
        histogram.reset()
        histogram.fill(
            self.results.values[other_loc_property], self.results.values[loc_property]
        )
        mesh = ax.pcolormesh(*histogram.axes.edges.T, histogram.view().T, **kwargs)
        fig.colorbar(mesh)  # type:ignore[union-attr]

        self.results.grouped.plot(
            kind="line",
            marker=".",
            x=None,
            y=self.parameter["loc_property"] + "_mean",
            color="blue",
            ax=ax,
        )

        if not self.results.grouped.expectation.isna().all():
            self.results.grouped.plot(
                kind="line",
                x=None,
                y="expectation",
                color="white",
                ax=ax,
            )

        ax.set(
            title="Grouped Property Expectation",
            xlabel=self.parameter["other_loc_property"],
            ylabel=self.parameter["loc_property"],
        )

        if log:
            ax.set(xscale="log", yscale="log")

        if fit:
            raise NotImplementedError

        return ax



"""

Utility methods for clustering locdata.

"""

from __future__ import annotations

import itertools
import sys
from collections.abc import Callable
from typing import Any

from locan.data.locdata import LocData

__all__: list[str] = ["serial_clustering"]


def serial_clustering(
    locdata: LocData,
    algorithm: Callable[..., Any],
    parameter_lists: dict[str, Any],
    **kwargs: Any,
) -> tuple[LocData, ...]:
    """
    Run and analyse a series of clustering processes to identify optimal
    parameters.

    Parameters
    ----------
    locdata
        Localization data.
    algorithm
        The locan clustering algorithm to use on locdata.
    parameter_lists
        A dictionary with all parameter lists that are to be iterated.
        The keys should be identical to parameter names of the used algorithm.
    kwargs
        Optional keyword arguments that are passed to the algorithm.

    Returns
    -------
    tuple[LocData, ...]
        The first element is a LocData object with a selection of all
        localizations that are defined as noise.
        If noise is false this element will be None.
        The second element is a new LocData instance assembling all generated
        selections (i.e. localization cluster).
    """
    parameter = locals()

    keys = parameter_lists.keys()
    products = list(itertools.product(*parameter_lists.values()))
    products_transposed = zip(*products)
    dictionary = {key: values for key, values in zip(keys, products_transposed)}
    results = []
    for product in products:
        iterated_arguments = {k: v for k, v in zip(keys, product)}
        results.append(algorithm(locdata, **iterated_arguments, **kwargs))

    noise_locdata = [result[0] for result in results]
    noise_collection = LocData.from_collection(noise_locdata)
    noise_collection.dataframe = noise_collection.dataframe.assign(**dictionary)
    collection_locdata = [result[1] for result in results]
    collection = LocData.from_collection(collection_locdata)
    collection.dataframe = collection.dataframe.assign(**dictionary)

    # metadata for noise_collection
    del noise_collection.meta.history[:]
    noise_collection.meta.history.add(
        name=sys._getframe().f_code.co_name, parameter=str(parameter)
    )

    # metadata for collection
    del collection.meta.history[:]
    collection.meta.history.add(
        name=sys._getframe().f_code.co_name, parameter=str(parameter)
    )

    return noise_collection, collection



import pytest

from locan.data.validation import _check_loc_properties


def test__check_loc_properties(locdata_2d):
    assert _check_loc_properties(locdata_2d) == locdata_2d.coordinate_keys
    assert _check_loc_properties(locdata_2d, loc_properties=["intensity"]) == [
        "intensity"
    ]
    with pytest.raises(ValueError):
        _check_loc_properties(locdata_2d, loc_properties=["undefined"])



"""
Benchmark functions for :mod:`locan.locan_io`
"""

import locan as lc


class LoadFiles:
    """
    Benchmarks for loading data from files
    """

    def setup(self):
        pass

    def time_load_asdf_file(self):
        file_path = lc.ROOT_DIR / "tests/test_data/npc_gp210.asdf"
        self.locdata = lc.load_asdf_file(path=file_path)

    def time_load_SMLM_file(self):
        file_path = lc.ROOT_DIR / "tests/test_data/SMLM_dstorm_data.smlm"
        self.locdata = lc.load_SMLM_file(path=file_path)

    def time_load_SMAP_file(self):
        file_path = lc.ROOT_DIR / "tests/test_data/smap_dstorm_data.mat"
        self.locdata = lc.load_SMAP_file(path=file_path)

    def time_load_txt_file(self):
        file_path = lc.ROOT_DIR / "tests/test_data/five_blobs_3D.txt"
        self.locdata = lc.load_txt_file(path=file_path)

    def time_load_decode_file(self):
        file_path = lc.ROOT_DIR / "tests/test_data/decode_dstorm_data.h5"
        self.locdata = lc.load_decode_file(path=file_path)


def main():
    bm = LoadFiles()
    bm.setup()
    bm.time_load_asdf_file()
    bm.time_load_SMLM_file()
    bm.time_load_SMAP_file()
    bm.time_load_txt_file()
    bm.time_load_decode_file()


def main_profile():
    """
    Run benchmarks through profiler.
    """
    import cProfile
    import pstats

    bm = LoadFiles()
    bm.setup()

    profiler = cProfile.Profile()
    profiler.enable()

    for _ in range(5):
        bm.time_load_asdf_file()
        bm.time_load_SMLM_file()
        bm.time_load_SMAP_file()
        bm.time_load_txt_file()
        bm.time_load_decode_file()

    profiler.disable()
    stats = pstats.Stats(profiler).sort_stats("cumtime")
    stats.print_stats("time_")


if __name__ == "__main__":
    # main()
    main_profile()



import matplotlib.pyplot as plt  # noqa: F401  # this import is needed for interactive tests
import numpy as np
import pytest

import locan
from locan import (  # noqa: F401  # this import is needed for interactive tests
    RenderEngine,
    render_2d_mpl,
    render_2d_napari,
    render_2d_napari_image,
    render_2d_rgb_napari,
    transform_affine,
)
from locan.dependencies import HAS_DEPENDENCY

napari = pytest.importorskip("napari")

HAS_NAPARI_AND_PYTESTQT = HAS_DEPENDENCY["napari"] and HAS_DEPENDENCY["pytestqt"]
# pytestqt is not a requested or extra dependency.
# If napari and pytest-qt is installed, all tests run.
# Tests in docker or GitHub actions on linux require xvfb
# for tests with pytest-qt to run.


@pytest.fixture()
def locdata_2d_negative_():
    """
    Fixture for returning `LocData` carrying 2D localizations including
    negative coordinates.
    """
    import pandas as pd

    locdata_dict = {
        "position_x": np.array([1, -1, 2, 3, 4, 5]) * 10,
        "position_y": np.array([1, 5, 3, 6, -2, 5]) * 10,
    }
    df = pd.DataFrame(locdata_dict)
    return locan.LocData.from_dataframe(dataframe=df)


@pytest.mark.visual
# this is to check overlay of rendered image and single localization points
def test_render_2d_napari_accurate_visual(locdata_2d_negative_):
    print(locdata_2d_negative_.data[locdata_2d_negative_.coordinate_keys].describe())

    render_2d_mpl(locdata_2d_negative_, bin_size=1, cmap="viridis")
    # plt.show()

    viewer = napari.Viewer()

    render_2d_napari(locdata_2d_negative_, bin_size=0.5, cmap="viridis", viewer=viewer)

    viewer.add_points(locdata_2d_negative_.coordinates, size=1, opacity=0.2)

    print(viewer.layers[0].corner_pixels)
    print(viewer.layers[0].data_to_world(viewer.layers[0].corner_pixels))

    napari.run()
    plt.close("all")


@pytest.mark.visual
# this is to check overlay of rendered image and single localization points
def test_render_2d_napari_new_visual(locdata_blobs_2d):
    print(locdata_blobs_2d.data[locdata_blobs_2d.coordinate_keys].describe())

    render_2d_mpl(locdata_blobs_2d, bin_size=10, cmap="viridis")
    plt.show()

    viewer = napari.Viewer()

    bins = locan.Bins(
        bin_size=(50, 100),
        bin_range=(
            (0, locdata_blobs_2d.coordinates.max(axis=0)[0]),
            (0, locdata_blobs_2d.coordinates.max(axis=0)[1]),
        ),
        extend_range=True,
    )

    render_2d_napari(locdata_blobs_2d, bins=bins, cmap="viridis", viewer=viewer)

    render_2d_napari(locdata_blobs_2d, bin_size=10, cmap="viridis", viewer=viewer)

    viewer.add_points(locdata_blobs_2d.coordinates, size=10, opacity=0.2)

    print(viewer.layers[0].corner_pixels)
    print(viewer.layers[0].data_to_world(viewer.layers[0].corner_pixels))

    napari.run()
    plt.close("all")


def test_render_2d_napari_image(locdata_blobs_2d):
    with pytest.raises(ValueError):
        # bins that are not equally sized cannot be displayed in napari.
        render_2d_napari_image(locdata_blobs_2d, bin_edges=((0, 10, 100), (0, 10, 100)))

    data, image_kwargs, layer_type = render_2d_napari_image(
        locdata_blobs_2d, bin_size=100, cmap="viridis", gamma=0.1
    )
    assert layer_type == "image"
    assert all(
        key in ["name", "colormap", "scale", "translate", "metadata", "gamma"]
        for key in image_kwargs
    )
    assert data.shape == (8, 5)


@pytest.mark.skipif(
    not HAS_NAPARI_AND_PYTESTQT, reason="Test requires napari and pytest-qt."
)
def test_render_2d_napari(make_napari_viewer, locdata_blobs_2d):
    viewer = make_napari_viewer()

    with pytest.raises(ValueError):
        # bins that are not equally sized cannot be displayed in napari.
        render_2d_napari(
            locdata_blobs_2d,
            viewer=viewer,
            cmap="viridis",
            bin_edges=((0, 10, 100), (0, 10, 100)),
        )

    viewer_ = render_2d_napari(
        locdata_blobs_2d, viewer=viewer, bin_size=100, cmap="viridis", gamma=0.1
    )
    assert viewer_ is viewer
    assert len(viewer.layers) == 1
    assert viewer.layers[0].name == "LocData 0"
    assert np.array_equal(viewer.layers[0].corner_pixels, [[0, 0], [7, 4]])
    # print(viewer.layers[0].data_to_world(viewer.layers[0].corner_pixels))
    assert np.array_equal(
        viewer.layers[0].data_to_world(viewer.layers[0].corner_pixels),
        [[162.0, 515.0], [862.0, 915.0]],
    )
    assert viewer.scale_bar.unit is None or len(viewer.scale_bar.unit) != 0
    assert viewer.layers[0].metadata["message"]


@pytest.mark.gui
def test_render_2d_napari_empty_gui(locdata_empty):
    render_2d_napari(locdata_empty, bin_size=100, cmap="viridis", gamma=0.1)
    napari.run()


@pytest.mark.skipif(
    not HAS_NAPARI_AND_PYTESTQT, reason="Test requires napari and pytest-qt."
)
def test_render_2d_napari_empty(make_napari_viewer, locdata_empty):
    viewer = make_napari_viewer()
    render_2d_napari(
        locdata_empty, viewer=viewer, bin_size=100, cmap="viridis", gamma=0.1
    )
    assert viewer.layers == []


@pytest.mark.gui
def test_render_2d_napari_single_gui(locdata_single_localization, caplog):
    render_2d_napari(
        locdata_single_localization, bin_size=100, cmap="viridis", gamma=0.1
    )
    assert caplog.record_tuples[-1] == (
        "locan.visualize.render_napari.render2d",
        30,
        "Locdata carries a single localization.",
    )
    napari.run()


@pytest.mark.skipif(
    not HAS_NAPARI_AND_PYTESTQT, reason="Test requires napari and pytest-qt."
)
def test_render_2d_napari_single(
    make_napari_viewer, locdata_single_localization, caplog
):
    viewer = make_napari_viewer()
    render_2d_napari(
        locdata_single_localization,
        viewer=viewer,
        bin_size=100,
        cmap="viridis",
        gamma=0.1,
    )
    assert caplog.record_tuples[-1] == (
        "locan.visualize.render_napari.render2d",
        30,
        "Locdata carries a single localization.",
    )


@pytest.mark.gui
def test_render_2d_rgb_napari_gui(locdata_blobs_2d):
    locdata_0 = locdata_blobs_2d
    locdata_1 = transform_affine(locdata_blobs_2d, offset=(20, 0))
    render_2d_rgb_napari([locdata_0, locdata_1], bin_size=20)
    napari.run()


@pytest.mark.skipif(
    not HAS_NAPARI_AND_PYTESTQT, reason="Test requires napari and pytest-qt."
)
def test_render_2d_rgb_napari(make_napari_viewer, locdata_blobs_2d):
    viewer = make_napari_viewer()
    locdata_0 = locdata_blobs_2d
    locdata_1 = transform_affine(locdata_blobs_2d, offset=(20, 0))
    render_2d_rgb_napari([locdata_0, locdata_1], viewer=viewer, bin_size=20)



"""

Compute localizations per frame.

"""

from __future__ import annotations

import logging
import sys
from collections.abc import Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

if TYPE_CHECKING:
    import matplotlib as mpl

    from locan.data.locdata import LocData

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import pandas as pd
from scipy import stats

from locan.analysis import metadata_analysis_pb2
from locan.analysis.analysis_base import _Analysis

__all__: list[str] = ["LocalizationsPerFrame"]

logger = logging.getLogger(__name__)


# The algorithms


def _localizations_per_frame(
    locdata: LocData | npt.ArrayLike,
    norm: int | float | str | None = None,
    time_delta: int | float | str | pd.Timedelta | None = "integration_time",
    resample: pd.DateOffset | pd.Timedelta | str | None = None,
    **kwargs: Any,
) -> pd.Series[Any]:
    """
    Compute localizations per frame.

    Parameters
    ----------
    locdata
       Points in time: either localization data that contains a column `frame`
       or an array with time points.
    norm
        Normalization factor that can be None, a number, or another property in
         `locdata`.
    time_delta
        Time per frame in milliseconds. String must specify the unit like
        "10ms".
         For "integration_time" the time is taken from
        locdata.meta.experiment.setups[0].optical_units[0].detection.camera.integration_time
    resample
        Parameter for :func:`pandas.Series.resample`: The offset string or
        object representing target conversion.
    kwargs
        Other parameters passed to :func:`pandas.Series.resample`.


    Returns:
    --------
    pandas.Series
    """
    # normalization
    if norm is None:
        normalization_factor: int | float = 1
        series_name = "n_localizations"
    elif isinstance(norm, str):
        try:
            normalization_factor = locdata.properties[norm]  # type: ignore
            series_name = "n_localizations / " + norm
        except (AttributeError, KeyError) as exception:
            raise KeyError(
                "normalization must be a valid property name in locdata.properties."
            ) from exception
    elif isinstance(norm, (int, float)):
        normalization_factor = norm
        series_name = f"n_localizations / {norm}"
    else:
        raise TypeError(
            "normalization should be None, a number or a valid property name."
        )

    try:
        frames_ = locdata.data.frame.astype(int)  # type: ignore
    except AttributeError:
        frames_ = np.asarray(locdata)

    frames, frame_counts = np.unique(frames_, return_counts=True)
    series = pd.Series(frame_counts, index=frames, dtype=float) / normalization_factor
    series.index.name = "frame"
    series.name = series_name

    if time_delta is None:
        pass
    elif isinstance(time_delta, pd.Timedelta):
        series.index = frames * time_delta
        series.index.name = "time"
    elif isinstance(time_delta, (int, float)):
        series.index = pd.to_timedelta(frames * time_delta, unit="ms")
        series.index.name = "time"
    elif time_delta == "integration_time":
        try:
            time_delta = (
                locdata.meta.experiment.setups[0]  # type: ignore
                .optical_units[0]
                .detection.camera.integration_time.ToTimedelta()
            )
            series.index = frames * time_delta
            series.index.name = "time"
        except (IndexError, AttributeError):
            logger.warning(
                "integration_time not available in locdata.meta - frames used instead."
            )
    elif isinstance(time_delta, str):
        time_delta = pd.Timedelta(time_delta)
        series.index = frames * time_delta
        series.index.name = "time"
    else:
        raise ValueError("The input for time_delta is not implemented.")

    if resample is not None:
        series = series.resample(resample, **kwargs).sum()

    return series


# The specific analysis classes


@dataclass(repr=False)
class _Results:
    time_series: pd.Series[Any]

    def accumulation_time(self, fraction: float = 0.5) -> int:
        normalized_cumulative_time_trace = (
            self.time_series.cumsum() / self.time_series.sum()
        )
        accumulation_time: int = normalized_cumulative_time_trace.gt(fraction).idxmax()
        return accumulation_time


class LocalizationsPerFrame(_Analysis):
    """
    Compute localizations per frame.

    Parameters
    ----------
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    norm : int | float | str | None
        Normalization factor that can be None, a number, or another property
        in `locdata`.
    time_delta : int | float | str | pd.Timedelta | None
        Time per frame in milliseconds. String must specify the unit like
        '10ms'.
        For 'integration_time' the time is taken from
        locdata.meta.experiment.setups[0].optical_units[0].detection.camera.integration_time
    resample : DateOffset | Timedelta | str
        Parameter for :func:`pandas.Series.resample`: The offset string or
        object representing target conversion.
    kwargs
        Other parameters passed to :func:`pandas.Series.resample`.

    Attributes
    ----------
    count : int
        A counter for counting instantiations.
    parameter : dict
        A dictionary with all settings for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    results : pandas.Series
        Computed results.
    distribution_statistics : Distribution_fits object | None
        Distribution parameters derived from MLE fitting of results.
    """

    count = 0

    def __init__(
        self,
        meta: metadata_analysis_pb2.AMetadata | None = None,
        norm: int | float | str | None = None,
        time_delta: int | float | str | pd.Timedelta | None = "integration_time",
        resample: pd.DateOffset | pd.Timedelta | str | None = None,
        **kwargs: Any,
    ) -> None:
        parameters = self._get_parameters(locals())
        super().__init__(**parameters)
        self.results = None
        self.distribution_statistics: _DistributionFits | None = None

    def compute(self, locdata: LocData | npt.ArrayLike) -> Self:
        """
        Run the computation.

        Parameters
        ----------
        locdata
           Points in time: either localization data that contains a column
           `frame` or an array with time points.

        Returns
        -------
        Self
        """
        if not len(locdata):  # type: ignore
            logger.warning("Locdata is empty.")
            self.distribution_statistics = None
            self.results = None
            return self

        self.distribution_statistics = None
        self.results = _Results(
            time_series=_localizations_per_frame(locdata=locdata, **self.parameter)
        )
        return self

    def fit_distributions(self, **kwargs: Any) -> None:
        # todo: fix kwarg
        """
        Fit probability density functions to the distributions of `
        loc_property` values in the results using MLE (scipy.stats).

        Parameters
        ----------
        loc_property : str
            The LocData property for which to fit an appropriate distribution;
            if None all plots are shown.
        """
        if self:
            self.distribution_statistics = _DistributionFits(self)
            self.distribution_statistics.fit(**kwargs)
        else:
            logger.warning("No results available to fit.")

    def plot(
        self,
        ax: mpl.axes.Axes | None = None,
        window: int = 1,
        cumulative: bool = False,
        normalize: bool = False,
        **kwargs: Any,
    ) -> mpl.axes.Axes:
        """
        Provide plot as :class:`matplotlib.axes.Axes` object showing the
        running average of results over window size.

        Parameters
        ----------
        ax
            The axes on which to show the image
        window
            Window for running average that is applied before plotting.
        cumulative
            Plot the cumulated results if true.
        normalize
            Normalize cumulative plot to the last value
        kwargs
            Other parameters passed to :func:`matplotlib.pyplot.plot`.

        Returns
        -------
        matplotlib.axes.Axes
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if self.results is None:
            return ax

        series = self.results.time_series

        # prepare plot
        if cumulative and normalize:
            _results = series.cumsum() / series.sum()
        elif cumulative and not normalize:
            _results = series.cumsum()
        elif not cumulative:
            _results = series
        else:
            return ax

        _results.rolling(window=window, center=True).mean().plot(ax=ax, **kwargs)

        ax.set(
            title=f"Localizations per Frame\n (window={window})",
            xlabel=series.index.name,
            ylabel=f"{series.name} (cumulative)" if cumulative else series.name,
        )

        return ax

    def hist(
        self,
        ax: mpl.axes.Axes | None = None,
        fit: bool = True,
        bins: int | Sequence[int | float] | str = "auto",
        **kwargs: Any,
    ) -> mpl.axes.Axes:
        """
        Provide histogram as :class:`matplotlib.axes.Axes` object showing
        hist(results).

        Parameters
        ----------
        ax
            The axes on which to show the image
        bins
            Bin specifications (passed to matplotlib.hist).
        fit
            Flag indicating if distributions fit are shown.
        kwargs
            Other parameters passed to :func:`matplotlib.pyplot.hist`.

        Returns
        -------
        matplotlib.axes.Axes
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if self.results is None:
            return ax

        series = self.results.time_series

        ax.hist(
            series.values, bins=bins, **dict(dict(density=True, log=False), **kwargs)
        )
        ax.set(title="Localizations per Frame", xlabel=series.name, ylabel="PDF")

        if fit:
            if self.distribution_statistics is None:
                self.fit_distributions()

            assert (  # type narrowing # noqa: S101
                self.distribution_statistics is not None
            )
            self.distribution_statistics.plot(ax=ax)

        return ax


# todo: add fit function


class _DistributionFits:
    """
    Handle for distribution fits.

    It holds the statistical parameters derived by fitting the result
    distributions using MLE (scipy.stats).
    Statistical parameters are defined as described in
    :ref:(https://docs.scipy.org/doc/scipy/reference/tutorial/stats/continuous.html)

    Parameters
    ----------
    analyis_class : LocalizationsPerFrame
        The analysis class with result data to fit.

    Attributes
    ----------
    analyis_class : LocalizationsPerFrame
        The analysis class with result data to fit.
    loc_property : str | None
        The LocData property for which to fit an appropriate distribution
    distribution : scipy.stats.rv_continuous | None
        Distribution model to fit.
    parameters : list[str]
    """

    def __init__(self, analysis_class: LocalizationsPerFrame) -> None:
        self.analysis_class = analysis_class
        if self.analysis_class.results is None:
            self.loc_property: str | None = None
        else:
            self.loc_property = self.analysis_class.results.time_series.name
        self.distribution: stats.rv_continuous | None = None
        self.parameters: list[str] = []

    def fit(
        self, distribution: stats.rv_continuous = stats.norm, **kwargs: Any
    ) -> None:
        """
        Fit scipy.stats.rv_continuous to analysis_class.results[loc_property].

        If with_constraints is true we put the following constraints on the fit
        procedure:
        If distribution is expon then
        floc=np.min(self.analysis_class.results[self.loc_property].values).

        Parameters
        ----------
        distribution
            Distribution model to fit.
        kwargs
            Other parameters are passed to
            :func:`scipy.stat.distribution.fit()`.
        """
        if self.analysis_class.results is None:
            return
        assert self.loc_property is not None  # type narrowing # noqa: S101

        self.distribution = distribution

        loc, scale = self.distribution.fit(
            self.analysis_class.results.time_series.values, **kwargs
        )
        self.parameters.extend(
            [self.loc_property + "_center", self.loc_property + "_sigma"]
        )
        setattr(self, self.loc_property + "_center", loc)
        setattr(self, self.loc_property + "_sigma", scale)

    def plot(self, ax: mpl.axes.Axes | None = None, **kwargs: Any) -> mpl.axes.Axes:
        """
        Provide plot as :class:`matplotlib.axes.Axes` object showing the
        probability distribution functions of fitted results.

        Parameters
        ----------
        ax
            The axes on which to show the image.
        kwargs
            Other parameters passed to :func:`matplotlib.pyplot.plot`.

        Returns
        -------
        matplotlib.axes.Axes
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if self.distribution is None:
            return ax

        # plot fit curve
        _center, _sigma = self.parameter_dict().values()

        x_values = np.linspace(
            self.distribution.ppf(0.001, loc=_center, scale=_sigma),
            self.distribution.ppf(0.999, loc=_center, scale=_sigma),
            100,
        )
        ax.plot(
            x_values,
            self.distribution.pdf(x_values, loc=_center, scale=_sigma),
            "r-",
            **dict(
                dict(lw=3, alpha=0.6, label=str(self.distribution.name) + " pdf"),
                **kwargs,
            ),
        )
        return ax

    def parameter_dict(self) -> dict[str, float]:
        """Dictionary of fitted parameters."""
        if self.parameters is None:
            return dict()
        else:
            return {k: self.__dict__[k] for k in self.parameters}



"""

Type definitions

"""

from __future__ import annotations

from collections.abc import Sequence
from typing import Any, Protocol, Union

import numpy as np

RandomGeneratorSeed = Union[
    None,
    int,
    Sequence[int],
    np.random.SeedSequence,
    np.random.BitGenerator,
    np.random.Generator,
]


class DataFrame(Protocol):
    """Dataframe that supports the dataframe interchange protocol."""

    def __dataframe__(
        self, nan_as_null: bool = False, allow_copy: bool = True
    ) -> DataFrame: ...


class LocData(Protocol):
    meta: Any
    references: Any

    @property
    def data(self) -> Any: ...

    def __len__(self) -> int: ...



"""
Benchmark functions for :mod:`locan.data.metadata_pb2`
"""

import locan as lc


class BenchmarkMetadata:
    """
    Benchmarks for data.metadata_pb2.Metadata objects
    """

    def setup(self):
        self.metadata = lc.data.metadata_pb2.Metadata()

    def time_initialize_metadata(self):
        lc.data.metadata_pb2.Metadata()

    def mem_metadata(self):
        return self.metadata


def main():
    bm = BenchmarkMetadata()
    bm.setup()
    bm.time_initialize_metadata()


def main_profile():
    """
    Run benchmarks through profiler.
    """
    import cProfile
    import pstats

    bm = BenchmarkMetadata()
    bm.setup()

    profiler = cProfile.Profile()
    profiler.enable()

    for _ in range(100):
        bm.time_initialize_metadata()

    profiler.disable()
    stats = pstats.Stats(profiler).sort_stats("cumtime")
    stats.print_stats("time_")
    stats.print_stats()


if __name__ == "__main__":
    # main()
    main_profile()



"""

File input/output for localization data Nanoimager files.

"""

from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from _typeshed import SupportsRead, SupportsReadline

import pandas as pd

import locan.constants
from locan.data import metadata_pb2
from locan.data.locdata import LocData
from locan.locan_io.locdata.utilities import (
    convert_property_names,
    convert_property_types,
    open_path_or_file_like,
)

__all__: list[str] = ["load_Nanoimager_header", "load_Nanoimager_file"]

logger = logging.getLogger(__name__)


def _read_Nanoimager_header(file: SupportsReadline[Any]) -> list[str]:
    """
    Read csv header from a Nanoimager single-molecule localization file and
    identify column names.

    Parameters
    ----------
    file
        Nanoimager file to read.

    Returns
    -------
    list[str]
        A list of valid dataset property keys as derived from the Nanoimager
        identifiers.
    """
    # read csv header
    header = file.readline().split("\n")[0]

    # list identifiers
    identifiers = [x.strip('"') for x in header.split(",")]

    column_keys = convert_property_names(
        properties=identifiers, property_mapping=locan.constants.NANOIMAGER_KEYS
    )
    return column_keys


def load_Nanoimager_header(
    path: str | os.PathLike[Any] | SupportsRead[Any],
) -> list[str]:
    """
    Load csv header from a Nanoimager single-molecule localization file and
    identify column names.

    Parameters
    ----------
    path
        File path for a Nanoimager file to load.

    Returns
    -------
    list[str]
        A list of valid dataset property keys as derived from the Nanoimager
        identifiers.
    """
    # read csv header
    with open_path_or_file_like(path) as file:
        return _read_Nanoimager_header(file)


def load_Nanoimager_file(
    path: str | os.PathLike[Any] | SupportsRead[Any],
    nrows: int | None = None,
    convert: bool = True,
    **kwargs: Any,
) -> LocData:
    """
    Load data from a Nanoimager single-molecule localization file.

    Parameters
    ----------
    path
        File path for a file to load.
    nrows
        The number of localizations to load from file. None means that all
        available rows are loaded.
    convert
        If True convert types by applying type specifications in
        locan.constants.PROPERTY_KEYS.
    kwargs
        Other parameters passed to `pandas.read_csv()`.

    Returns
    -------
    LocData
        A new instance of LocData with all localizations.
    """
    with open_path_or_file_like(path) as file:
        columns = _read_Nanoimager_header(file)
        dataframe = pd.read_csv(
            file, sep=",", skiprows=0, nrows=nrows, names=columns, **kwargs
        )

    if convert:
        dataframe = convert_property_types(
            dataframe, types=locan.constants.PROPERTY_KEYS
        )

    dat = LocData.from_dataframe(dataframe=dataframe)

    dat.meta.source = metadata_pb2.EXPERIMENT
    dat.meta.state = metadata_pb2.RAW
    dat.meta.file.type = metadata_pb2.NANOIMAGER
    dat.meta.file.path = str(path)

    for property_ in sorted(
        list(set(columns).intersection({"position_x", "position_y", "position_z"}))
    ):
        dat.meta.localization_properties.add(name=property_, unit="nm", type="float")

    del dat.meta.history[:]
    dat.meta.history.add(
        name="load_Nanoimager_file",
        parameter=f"path={str(path)}, nrows={nrows}",
    )

    return dat



import logging
from copy import deepcopy

import numpy as np
import pytest

from locan import transform_counts_to_photons
from locan.process.transform.intensity_transformation import (
    _transform_counts_to_photons,
)


def test__transform_counts_to_photons():
    intensities = np.arange(10, 20)
    new_intensities = _transform_counts_to_photons(
        intensities=intensities, offset=10, gain=10, electrons_per_count=10
    )
    assert np.array_equal(new_intensities, np.arange(0, 10))

    intensities = list(range(10, 20))
    new_intensities = _transform_counts_to_photons(
        intensities=intensities, offset=10, gain=10, electrons_per_count=10
    )
    assert np.array_equal(new_intensities, np.arange(0, 10))


def test_transform_counts_to_photons(locdata_2d, caplog):
    caplog.set_level(logging.INFO)
    locdata = deepcopy(locdata_2d)
    df = locdata.dataframe.assign(local_background=[100] * len(locdata))
    locdata.update(dataframe=df)

    # metadata not available
    with pytest.raises(IndexError):
        transform_counts_to_photons(locdata=locdata)

    # provide camera metadata
    setup = locdata.meta.experiment.setups.add()
    ou = setup.optical_units.add()
    ou.detection.camera.offset = 10
    ou.detection.camera.gain = 10
    ou.detection.camera.electrons_per_count = 2

    # test
    new_locdata = transform_counts_to_photons(locdata=locdata)
    assert np.array_equal(
        new_locdata.data.intensity, [18.0, 28.0, 20.0, 14.0, 19.0, 17.0]
    )
    index = [prob.name for prob in new_locdata.meta.localization_properties].index(
        "intensity"
    )
    assert new_locdata.meta.localization_properties[index].unit == "photons"

    new_locdata = transform_counts_to_photons(
        locdata=locdata, loc_properties="local_background"
    )
    assert np.array_equal(new_locdata.data.local_background, [18.0] * len(new_locdata))
    index = [prob.name for prob in new_locdata.meta.localization_properties].index(
        "local_background"
    )
    assert new_locdata.meta.localization_properties[index].unit == "photons"

    new_locdata = transform_counts_to_photons(
        locdata=locdata, loc_properties=["intensity", "local_background"]
    )
    assert np.array_equal(
        new_locdata.data.intensity, [18.0, 28.0, 20.0, 14.0, 19.0, 17.0]
    )
    index = [prob.name for prob in new_locdata.meta.localization_properties].index(
        "intensity"
    )
    assert new_locdata.meta.localization_properties[index].unit == "photons"
    assert np.array_equal(new_locdata.data.local_background, [18.0] * len(new_locdata))
    index = [prob.name for prob in new_locdata.meta.localization_properties].index(
        "local_background"
    )
    assert new_locdata.meta.localization_properties[index].unit == "photons"

    new_locdata = transform_counts_to_photons(
        locdata=locdata,
        metadata=locdata.meta.experiment.setups[0].optical_units[0].detection.camera,
    )
    assert np.array_equal(
        new_locdata.data.intensity, [18.0, 28.0, 20.0, 14.0, 19.0, 17.0]
    )
    index = [prob.name for prob in new_locdata.meta.localization_properties].index(
        "intensity"
    )
    assert new_locdata.meta.localization_properties[index].unit == "photons"

    new_locdata = transform_counts_to_photons(
        locdata=locdata, loc_properties=["intensity", "not_present", "local_background"]
    )
    assert caplog.record_tuples[-2] == (
        "locan.process.transform.intensity_transformation",
        30,
        "Localization property not_present is not available.",
    )
    assert caplog.record_tuples[-1] == (
        "locan.process.transform.intensity_transformation",
        20,
        "Successfully converted: ['intensity', 'local_background'].",
    )

    new_locdata = transform_counts_to_photons(locdata=locdata)
    new_locdata_2 = transform_counts_to_photons(locdata=new_locdata)
    assert np.array_equal(new_locdata_2.data.intensity, new_locdata.data.intensity)
    assert caplog.record_tuples[-2] == (
        "locan.process.transform.intensity_transformation",
        30,
        "Localization property intensity is already provided with unit photons",
    )



"""

Register localization data.

This module registers localization data and provides transformation parameters
to put other localization data in registry.

Parts of this code is adapted from https://github.com/jungmannlab/picasso.
(MIT license, Copyright (c) 2016 Jungmann Lab, MPI of Biochemistry)
"""

from __future__ import annotations

import warnings
from collections.abc import Sequence
from functools import wraps
from typing import TYPE_CHECKING, Any, Literal, NamedTuple

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
from lmfit import Model, Parameters
from skimage.registration import phase_cross_correlation

from locan.data.locdata import LocData
from locan.dependencies import HAS_DEPENDENCY, needs_package
from locan.process.aggregate import Bins, histogram
from locan.process.properties.locdata_statistics import range_from_collection
from locan.process.transform.spatial_transformation import _homogeneous_matrix

if HAS_DEPENDENCY["open3d"]:
    import open3d as o3d

if TYPE_CHECKING:
    import boost_histogram as bh


__all__: list[str] = ["register_icp", "register_cc"]


class Transformation(NamedTuple):
    matrix: npt.NDArray[np.float64]
    offset: npt.NDArray[np.float64]


@needs_package("open3d")
def _register_icp_open3d(
    points: npt.ArrayLike,
    other_points: npt.ArrayLike,
    matrix: npt.ArrayLike | None = None,
    offset: npt.ArrayLike | None = None,
    pre_translation: npt.ArrayLike | None = None,
    max_correspondence_distance: float = 1_000,
    max_iteration: int = 10_000,
    with_scaling: bool = True,
    verbose: bool = True,
) -> Transformation:
    """
    Register `points` by an "Iterative Closest Point" algorithm using open3d.

    Parameters
    ----------
    points
        Points representing the source on which to perform the manipulation.
    other_points
        Points representing the target.
    matrix
        Transformation matrix with shape (d, d)used as initial value.
        If None the unit matrix is used.
    offset
        Translation vector with shape (d,) used as initial value.
        If None a vector of zeros is used.
    pre_translation
        Values for translation of coordinates before registration.
    max_correspondence_distance
        Threshold distance for the icp algorithm. Parameter is passed to
        open3d algorithm.
    max_iteration
        Maximum number of iterations. Parameter is passed to open3d algorithm.
    with_scaling
        Allow scaling transformation. Parameter is passed to open3d algorithm.
    verbose
        Flag indicating if transformation results are printed out.

    Returns
    -------
    Transformation
        Matrix and offset representing the optimized transformation.
    """
    points_ = np.asarray(points)
    other_points_ = np.asarray(other_points)

    # prepare 3d
    if np.shape(points)[1] == np.shape(other_points)[1]:
        dimension = np.shape(points)[1]
    else:
        raise ValueError("Dimensions for locdata and other_locdata are incompatible.")

    if dimension == 2:
        points_3d = np.concatenate([points_, np.zeros((len(points_), 1))], axis=1)
        other_points_3d = np.concatenate(
            [other_points_, np.zeros((len(other_points_), 1))], axis=1
        )
    elif dimension == 3:
        points_3d = points_
        other_points_3d = other_points_
    else:
        raise ValueError("Point array has the wrong shape.")

    # points in open3d
    point_cloud = o3d.geometry.PointCloud()
    point_cloud.points = o3d.utility.Vector3dVector(points_3d)

    other_point_cloud = o3d.geometry.PointCloud()
    other_point_cloud.points = o3d.utility.Vector3dVector(other_points_3d)

    # initial matrix
    matrix_ = np.identity(dimension) if matrix is None else np.asarray(matrix)
    offset_ = np.zeros(dimension) if offset is None else np.asarray(offset)

    matrix_3d = np.identity(3)
    matrix_3d[:dimension, :dimension] = matrix_
    offset_3d = np.zeros(3)
    offset_3d[:dimension] = offset_
    matrix_homogeneous = _homogeneous_matrix(matrix_3d, offset_3d)

    if pre_translation is not None:
        pre_translation_3d = np.zeros(3)
        pre_translation_3d[:dimension] = pre_translation
        other_point_cloud.translate(pre_translation_3d)

    # apply ICP
    registration = o3d.pipelines.registration.registration_icp(
        source=point_cloud,
        target=other_point_cloud,
        max_correspondence_distance=max_correspondence_distance,
        init=matrix_homogeneous,
        estimation_method=o3d.pipelines.registration.TransformationEstimationPointToPoint(
            with_scaling=with_scaling
        ),
        criteria=o3d.pipelines.registration.ICPConvergenceCriteria(
            max_iteration=max_iteration
        ),
    )

    if dimension == 2:
        new_matrix = registration.transformation[0:2, 0:2]
        new_offset = registration.transformation[0:2, 3]
    else:  # if dimension == 3:
        new_matrix = registration.transformation[0:3, 0:3]
        new_offset = registration.transformation[0:3, 3]

    if verbose:
        print(registration)

    return Transformation(new_matrix, new_offset)


def register_icp(
    locdata: npt.ArrayLike | LocData,
    other_locdata: npt.ArrayLike | LocData,
    matrix: npt.ArrayLike | None = None,
    offset: npt.ArrayLike | None = None,
    pre_translation: npt.ArrayLike | None = None,
    max_correspondence_distance: float = 1_000,
    max_iteration: int = 10_000,
    verbose: bool = True,
) -> Transformation:
    """
    Register `points` or coordinates in `locdata` by an
    "Iterative Closest Point" algorithm using open3d.

    Parameters
    ----------
    locdata
        Localization data representing the source on which to perform the
        manipulation.
    other_locdata
        Localization data representing the target.
    matrix
        Transformation matrix with shape (d, d) used as initial value.
        If None the unit matrix is used.
    offset
        Translation vector with shape (d,) used as initial value.
        If None a vector of zeros is used.
    pre_translation
        Values for translation of coordinates before registration.
    max_correspondence_distance
        Threshold distance for the icp algorithm. Parameter is passed to
        open3d algorithm.
    max_iteration
        Maximum number of iterations. Parameter is passed to open3d algorithm.
    verbose
        Flag indicating if transformation results are printed out.

    Returns
    -------
    Transformation
        Matrix and offset representing the optimized transformation.
    """
    local_parameter = locals()

    # adjust input
    if isinstance(locdata, LocData):
        points = locdata.coordinates
    else:
        points = np.asarray(locdata)

    if isinstance(other_locdata, LocData):
        other_points = other_locdata.coordinates
    else:
        other_points = np.asarray(other_locdata)

    transformation = _register_icp_open3d(
        points,
        other_points,
        matrix=matrix,
        offset=offset,
        pre_translation=pre_translation,
        max_correspondence_distance=1_000,
        max_iteration=10_000,
        verbose=True,
    )
    return transformation


def _xcorr(imageA: npt.ArrayLike, imageB: npt.ArrayLike) -> npt.NDArray[np.float64]:
    """
    This function is adapted from picasso/imageprocess
    by Joerg Schnitzbauer, MPI of Biochemistry
    https://github.com/jungmannlab/picasso/blob/master/picasso/imageprocess.py
    (MIT license, Copyright (c) 2016 Jungmann Lab, MPI of Biochemistry)
    """
    imageA = np.asarray(imageA)
    imageB = np.asarray(imageB)
    FimageA = np.fft.fft2(imageA)
    CFimageB = np.conj(np.fft.fft2(imageB))
    return_value: npt.NDArray[np.float64] = np.fft.fftshift(
        np.real(np.fft.ifft2(FimageA * CFimageB))
    ) / np.sqrt(imageA.size)
    return return_value


def _get_image_shift(
    imageA: npt.ArrayLike,
    imageB: npt.ArrayLike,
    box: Any,
    roi: Any | None = None,
    display: bool = False,
) -> tuple[int | float, int | float]:
    """
    Computes the shift from imageA to imageB.

    This function is adapted from picasso/imageprocess
    by Joerg Schnitzbauer, MPI of Biochemistry
    https://github.com/jungmannlab/picasso/blob/master/picasso/imageprocess.py
    (MIT license, Copyright (c) 2016 Jungmann Lab, MPI of Biochemistry)
    """
    imageA = np.asarray(imageA)
    imageB = np.asarray(imageB)
    if (np.sum(imageA) == 0) or (np.sum(imageB) == 0):
        return 0, 0
    # Compute image correlation
    XCorr = _xcorr(imageA, imageB)
    # Cut out center roi
    Y, X = imageA.shape
    if roi is not None:
        Y_ = int((Y - roi) / 2)
        X_ = int((X - roi) / 2)
        if Y_ > 0:
            XCorr = XCorr[Y_:-Y_, :]
        else:
            Y_ = 0
        if X_ > 0:
            XCorr = XCorr[:, X_:-X_]
        else:
            X_ = 0
    else:
        Y_ = X_ = 0
    # A quarter of the fit ROI
    fit_X = int(box / 2)
    # A coordinate grid for the fitting ROI
    y, x = np.mgrid[-fit_X : fit_X + 1, -fit_X : fit_X + 1]
    # Find the brightest pixel and cut out the fit ROI
    y_max_, x_max_ = np.unravel_index(XCorr.argmax(), XCorr.shape)
    FitROI = XCorr[
        y_max_ - fit_X : y_max_ + fit_X + 1,
        x_max_ - fit_X : x_max_ + fit_X + 1,
    ]

    dimensions = FitROI.shape

    if 0 in dimensions or dimensions[0] != dimensions[1]:
        xc, yc = 0, 0
    else:
        # The fit model based on lmfit
        def flat_2d_gaussian(a, xc, yc, s, b):  # type: ignore
            A = a * np.exp(-0.5 * ((x - xc) ** 2 + (y - yc) ** 2) / s**2) + b
            return A.flatten()

        gaussian2d = Model(flat_2d_gaussian, name="2D Gaussian", independent_vars=[])

        # Set up initial parameters and fit
        params = Parameters()
        params.add("a", value=FitROI.max(), vary=True, min=0)
        params.add("xc", value=0, vary=True)
        params.add("yc", value=0, vary=True)
        params.add("s", value=1, vary=True, min=0)
        params.add("b", value=FitROI.min(), vary=True, min=0)
        results = gaussian2d.fit(FitROI.flatten(), params)

        # Get maximum coordinates and add offsets
        xc = results.best_values["xc"]
        yc = results.best_values["yc"]
        xc += X_ + x_max_  # type: ignore
        yc += Y_ + y_max_  # type: ignore

        if display:
            plt.figure(figsize=(17, 10))
            plt.subplot(1, 3, 1)  # type: ignore
            plt.imshow(imageA, interpolation="none")
            plt.subplot(1, 3, 2)  # type: ignore
            plt.imshow(imageB, interpolation="none")
            plt.subplot(1, 3, 3)  # type: ignore
            plt.imshow(XCorr, interpolation="none")
            plt.plot(xc, yc, "x")
            plt.show()  # type: ignore

        xc -= np.floor(X / 2)
        yc -= np.floor(Y / 2)

    return -yc, -xc


def _register_cc_picasso(
    locdata: npt.ArrayLike | LocData,
    other_locdata: npt.ArrayLike | LocData,
    max_offset: int | float | None = None,
    bins: Bins | bh.axis.Axis | bh.axis.AxesTuple | None = None,
    n_bins: int | Sequence[int] | None = None,
    bin_size: float | Sequence[float] | Sequence[Sequence[float]] | None = None,
    bin_edges: Sequence[float] | Sequence[Sequence[float]] | None = None,
    bin_range: (
        tuple[float, float]
        | Sequence[float]
        | Sequence[Sequence[float]]
        | Literal["zero", "link"]
        | None
    ) = None,
    verbose: bool = False,
) -> Transformation:
    """
    Register `points` or coordinates in `locdata` by a
    cross-correlation algorithm.

    This function is based on code from picasso/imageprocess
    by Joerg Schnitzbauer, MPI of Biochemistry
    https://github.com/jungmannlab/picasso/blob/master/picasso/imageprocess.py

    Parameters
    ----------
    locdata
        Localization data representing the source on which to perform the
        manipulation.
    other_locdata
        Localization data representing the target.
    max_offset
        Maximum possible offset.
    bins
        Specific class specifying the bins.
    bin_edges
        Bin edges for all or each dimension
        with shape (dimension, n_bin_edges).
    bin_range
        Minimum and maximum edge for all or each dimensions
        with shape (2,) or (dimension, 2).
        If None (min, max) ranges are determined from data and returned;
        if 'zero' (0, max) ranges with max determined from data are returned.
        if 'link' (min_all, max_all) ranges with min and max determined from
        all combined data are returned.
    n_bins
        The number of bins for all or each dimension.
        5 yields 5 bins in all dimensions.
        (2, 5) yields 2 bins for one dimension and 5 for the other dimension.
    bin_size
        The size of bins for all or each bin and for all or each dimension
        with shape (dimension,) or (dimension, n_bins).
        5 would describe bin_size of 5 for all bins in all dimensions.
        ((2, 5),) yield bins of size (2, 5) for one dimension.
        (2, 5) yields bins of size 2 for one dimension and 5 for the other
        dimension.
        ((2, 5), (1, 3)) yields bins of size (2, 5) for one dimension and
        (1, 3) for the other dimension.
        To specify arbitrary sequence of `bin_size` use `bin_edges` instead.
    verbose
        Flag indicating if transformation results are printed out.

    Returns
    -------
    Transformation
        Matrix and offset representing the optimized transformation.
    """
    warnings.warn(
        "The function _register_cc_picasso is deprecated."
        "Use _register_cc_skimage instead.",
        DeprecationWarning,
        stacklevel=2,
    )

    if isinstance(locdata, LocData) and isinstance(other_locdata, LocData):
        bin_range_: (
            tuple[float, float]
            | Sequence[float]
            | Sequence[Sequence[float]]
            | Literal["zero", "link"]
            | None
        )
        if bin_range is None:
            bin_range_ = range_from_collection([locdata, other_locdata])
        else:
            bin_range_ = bin_range

        image, _, _ = histogram(
            locdata,
            bins=bins,
            n_bins=n_bins,
            bin_size=bin_size,
            bin_edges=bin_edges,
            bin_range=bin_range_,
        )
        other_image, _, _ = histogram(
            other_locdata,
            bins=bins,
            n_bins=n_bins,
            bin_size=bin_size,
            bin_edges=bin_edges,
            bin_range=bin_range_,
        )

    else:
        image = np.asarray(locdata)
        other_image = np.asarray(other_locdata)

    dimension = image.ndim
    matrix = np.identity(dimension)
    # todo: turn box into parameter
    offset = _get_image_shift(
        image, other_image, box=5, roi=max_offset, display=verbose
    )
    offset_ = np.asarray(offset) * bin_size

    return Transformation(matrix, offset_)


def _register_cc_skimage(
    locdata: npt.ArrayLike | LocData,
    other_locdata: npt.ArrayLike | LocData,
    bins: Bins | bh.axis.Axis | bh.axis.AxesTuple | None = None,
    n_bins: int | Sequence[int] | None = None,
    bin_size: float | Sequence[float] | Sequence[Sequence[float]] | None = None,
    bin_edges: Sequence[float] | Sequence[Sequence[float]] | None = None,
    bin_range: (
        tuple[float, float]
        | Sequence[float]
        | Sequence[Sequence[float]]
        | Literal["zero", "link"]
        | None
    ) = None,
    **kwargs: Any,
) -> Transformation:
    """
    Register `points` or coordinates in `locdata` by a
    cross-correlation algorithm.

    Parameters
    ----------
    locdata
        Localization data representing the source on which to perform the
        manipulation.
    other_locdata
        Localization data representing the target.
    bins
        Specific class specifying the bins.
    bin_edges
        Bin edges for all or each dimension
        with shape (dimension, n_bin_edges).
    bin_range
        Minimum and maximum edge for all or each dimensions
        with shape (2,) or (dimension, 2).
        If None (min, max) ranges are determined from data and returned;
        if 'zero' (0, max) ranges with max determined from data are returned.
        if 'link' (min_all, max_all) ranges with min and max determined from
        all combined data are returned.
    n_bins
        The number of bins for all or each dimension.
        5 yields 5 bins in all dimensions.
        (2, 5) yields 2 bins for one dimension and 5 for the other dimension.
    bin_size
        The size of bins for all or each bin and for all or each dimension
        with shape (dimension,) or (dimension, n_bins).
        5 would describe bin_size of 5 for all bins in all dimensions.
        ((2, 5),) yield bins of size (2, 5) for one dimension.
        (2, 5) yields bins of size 2 for one dimension and 5 for the other
        dimension.
        ((2, 5), (1, 3)) yields bins of size (2, 5) for one dimension and
        (1, 3) for the other dimension.
        To specify arbitrary sequence of `bin_size` use `bin_edges` instead.
    kwargs
        Other parameters passed to :func:`skimage.phase_cross_correlation`.

    Returns
    -------
    Transformation
        Matrix and offset representing the optimized transformation.
    """
    if isinstance(locdata, LocData) and isinstance(other_locdata, LocData):
        bin_range_: (
            tuple[float, float]
            | Sequence[float]
            | Sequence[Sequence[float]]
            | Literal["zero", "link"]
            | None
        )
        if bin_range is None:
            bin_range_ = range_from_collection([locdata, other_locdata])
        else:
            bin_range_ = bin_range

        image, _, _ = histogram(
            locdata,
            bins=bins,
            n_bins=n_bins,
            bin_size=bin_size,
            bin_edges=bin_edges,
            bin_range=bin_range_,
        )
        other_image, _, _ = histogram(
            other_locdata,
            bins=bins,
            n_bins=n_bins,
            bin_size=bin_size,
            bin_edges=bin_edges,
            bin_range=bin_range_,
        )

    else:
        image = np.asarray(locdata)
        other_image = np.asarray(other_locdata)

    dimension = image.ndim
    matrix = np.identity(dimension)

    shift, _, _ = phase_cross_correlation(  # type: ignore
        reference_image=other_image,
        moving_image=image,
        **dict(
            dict(
                upsample_factor=1,
                disambiguate=False,
                normalization=None,  # or 'phase'
            ),
            **kwargs,
        ),
    )
    offset_ = np.asarray(shift) * bin_size

    return Transformation(matrix, offset_)


@wraps(_register_cc_skimage)
def register_cc(*args: Any, **kwargs: Any) -> Transformation:
    warnings.warn(
        "The function register_cc has been refactored. "
        "The kwargs max_offset and verbose are deprecated . "
        "It now calls _register_cc_skimage. "
        "Use _register_cc_picasso for legacy behavior.",
        UserWarning,
        stacklevel=2,
    )
    return _register_cc_skimage(*args, **kwargs)



from io import StringIO

import locan.constants
from locan import FileType
from locan.locan_io import load_Nanoimager_file
from locan.locan_io.locdata.nanoimager_io import load_Nanoimager_header


def test_get_correct_column_names_from_Nanoimager_header():
    columns = load_Nanoimager_header(
        path=locan.ROOT_DIR / "tests/test_data/Nanoimager_dstorm_data.csv"
    )
    assert columns == [
        "channel",
        "frame",
        "position_x",
        "position_y",
        "position_z",
        "intensity",
        "local_background",
    ]

    file_like = StringIO(
        "Channel,Frame,X (nm),Y (nm),Z (nm),Photons,Background\n"
        "0,1548,40918.949219,56104.691406,0.000000,139.828232,0.848500"
    )
    columns = load_Nanoimager_header(path=file_like)
    assert columns == [
        "channel",
        "frame",
        "position_x",
        "position_y",
        "position_z",
        "intensity",
        "local_background",
    ]


def test_loading_Nanoimager_file():
    file_path = locan.ROOT_DIR / "tests/test_data/Nanoimager_dstorm_data.csv"
    dat = load_Nanoimager_file(path=file_path, nrows=10)
    # print(dat.data.columns)
    assert len(dat) == 10
    assert all(
        dat.data.columns
        == [
            "channel",
            "frame",
            "position_x",
            "position_y",
            "position_z",
            "intensity",
            "local_background",
        ]
    )
    assert dat.meta.file.type == FileType.NANOIMAGER.value
    assert dat.meta.file.path == str(file_path)

    file_like = StringIO(
        "Channel,Frame,X (nm),Y (nm),Z (nm),Photons,Background\n"
        "0,1548,40918.949219,56104.691406,0.000000,139.828232,0.848500"
    )
    dat = load_Nanoimager_file(path=file_like, nrows=1)
    assert len(dat) == 1
    assert all(
        dat.data.columns
        == [
            "channel",
            "frame",
            "position_x",
            "position_y",
            "position_z",
            "intensity",
            "local_background",
        ]
    )
